from __future__ import absolute_import
from __future__ import print_function

import os
from collections import namedtuple, MutableMapping
from copy import deepcopy, copy
from itertools import ifilter
import logging

import pyros_utils
import rospy
import rosservice, rostopic, rosparam

import re
import ast
import socket
import threading
import Queue

import time

# create logger
_logger = logging.getLogger(__name__)
# and let it propagate to parent logger, or other handler
# the user of pyros should configure handlers

from ..baseinterface import DiffTuple
from .baseinterface import BaseInterface
from .param_if_pool import RosParamIfPool
from .service_if_pool import RosServiceIfPool
from .subscriber_if_pool import RosSubscriberIfPool
from .publisher_if_pool import RosPublisherIfPool

from .connection_cache_utils import connection_cache_proxy_create, connection_cache_marshall, connection_cache_merge_marshalled

try:
    import rocon_python_comms
except ImportError:
    rocon_python_comms = None

# To make sure things dont get messed up between threads
CacheTuple = (namedtuple("CacheTuple", "complete added removed"))


class RosInterface(BaseInterface):

    """
    RosInterface.
    """
    def __init__(self, node_name, publishers=None, subscribers=None, services=None, params=None, enable_cache=False, argv=None):
        # This runs in a child process (managed by PyrosROS) and as a normal ros node)

        # First thing to do : find the rosmaster...
        # master has to be running here or we just wait for ever
        # TODO : improve...
        m, _ = pyros_utils.get_master(spawn=False)
        while not m.is_online():
            _logger.warning("ROSMASTER not found !!!...")
            time.sleep(0.5)

        # Second thing to do : initialize the ros node and disable signals to avoid overriding callers behavior
        rospy.init_node(node_name, argv=argv, disable_signals=True)
        rospy.loginfo('RosInterface {name} node started with args : {argv}'.format(name=node_name, argv=argv))
        # note on dynamic update (config reload, etc.) this is reinitialized
        # rospy.init_node supports being reinitialized with the exact same arguments

        self.enable_cache = enable_cache
        if self.enable_cache and rocon_python_comms is None:
            rospy.logerr("Connection Cache enabled for RosInterface, but rocon_python_comms not found. Disabling.")
            self.enable_cache = False

        # only one queue to avoid sync issues
        self.cb_ss = Queue.Queue()

        # we add params from ROS environment, if we get something there (bwcompat behavior)
        services = services or []
        publishers = publishers or []
        subscribers = subscribers or []
        params = params or []
        services += list(set(ast.literal_eval(rospy.get_param('~services', "[]"))))

        # bwcompat
        publishers += list(set(ast.literal_eval(rospy.get_param('~topics', "[]"))))
        subscribers += list(set(ast.literal_eval(rospy.get_param('~topics', "[]"))))
        # new
        publishers += list(set(ast.literal_eval(rospy.get_param('~publishers', "[]"))))
        subscribers += list(set(ast.literal_eval(rospy.get_param('~subscribers', "[]"))))

        params += list(set(ast.literal_eval(rospy.get_param('~params', "[]"))))
        enable_cache = rospy.get_param('~enable_cache', enable_cache)

        if enable_cache is not None:
            self.enable_cache = enable_cache
        # Note : None means no change ( different from [] )
        rospy.loginfo("""[{name}] ROS Interface initialized with:
        -    services : {services}
        -    publishers : {publishers}
        -    subscribers : {subscribers}
        -    params : {params}
        -    enable_cache : {enable_cache}
        """.format(
            name=__name__,
            publishers="\n" + "- ".rjust(10) + "\n\t- ".join(publishers) if publishers else [],
            subscribers="\n" + "- ".rjust(10) + "\n\t- ".join(subscribers) if subscribers else [],
            services="\n" + "- ".rjust(10) + "\n\t- ".join(services) if services else [],
            params="\n" + "- ".rjust(10) + "\n\t- ".join(params) if params else [],
            enable_cache=enable_cache)
        )

        # This base constructor assumes the system to interface with is already available ( can do a get_svc_available() )
        params_pool = RosParamIfPool(params)
        services_pool = RosServiceIfPool(services)
        subscribers_pool = RosSubscriberIfPool(subscribers)
        publishers_pool = RosPublisherIfPool(publishers)

        super(RosInterface, self).__init__(publishers_pool, subscribers_pool, services_pool, params_pool)

        # connecting to the master via proxy object
        self._master = rospy.get_master()

        #: If enabled, connection cache proxy will be setup in update() to allow dynamic update via config.
        # TODO : double check : maybe useless now since we completely reinit the interface for dynamic update...
        self.connection_cache = None

        # Setup our debug log
        # We need this debug log since rospy.logdebug does NOT store debug messages in the log.
        # But it should Ref : http://wiki.ros.org/rospy/Overview/Logging#Reading_log_messages
        # TODO : find why.
        ros_home = os.environ.get('ROS_HOME', os.path.join(os.path.expanduser("~"), '.ros'))
        self._debug_logger = logging.getLogger('pyros.ros_interface')
        if not os.path.exists(os.path.join(ros_home, 'logdebug')):
            os.makedirs(os.path.join(ros_home, 'logdebug'))
        logfilename = os.path.join(ros_home, 'logdebug', rospy.get_name()[1:].replace(os.path.sep, "-") + '_pyros_rosinterface.log')
        file_handler = logging.handlers.RotatingFileHandler(
            logfilename,
            maxBytes=100 * 131072,
            backupCount=10
        )
        self._debug_logger.propagate = False  # to avoid propagating to root logger
        self._debug_logger.setLevel(logging.DEBUG)
        self._debug_logger.addHandler(file_handler)

    def retrieve_system_state(self):
        """
        This will retrieve the system state from ROS master if needed, and apply changes to local variable to keep
        a local representation of the connections available up to date.
        """
        try:
            # we call the master only if we dont get system_state from connection cache
            if self.enable_cache and self.connection_cache is not None:
                publishers, subscribers, services = self.connection_cache.getSystemState()
                topic_types = self.connection_cache.getTopicTypes()
                try:
                    service_types = self.connection_cache.getServiceTypes()
                    # handling fallback here since master doesnt have the API
                except rocon_python_comms.UnknownSystemState as exc:
                    service_types = []
            else:
                publishers, subscribers, services = self._master.getSystemState()[2]
                topic_types = self._master.getTopicTypes()[2]
                service_types = []  # master misses this API to be consistent

            # Getting this doesnt depend on cache for now
            params = set(rospy.get_param_names())

            return publishers, subscribers, services, params, topic_types, service_types

        except socket.error:
            rospy.logerr("[{name}] couldn't get system state from the master ".format(name=__name__))

    # for use with line_profiler or memory_profiler
    # Not working yet... need to solve multiprocess profiling issues...
    #@profile
    def update(self):

        backedup_complete_cb_ss = None

        #update will retrieve system state here
        publishers = []
        subscribers = []
        services = []
        params = []
        topic_types = []
        service_types = []

        # and populate these to represent the changes done on the interface
        params_if_dt = DiffTuple([], [])
        services_if_dt = DiffTuple([], [])
        subscribers_if_dt = DiffTuple([], [])
        publishers_if_dt = DiffTuple([], [])

        # Destroying connection cache proxy if needed
        if self.connection_cache is not None and not self.enable_cache:
            # removing existing connection cache proxy to force a reinit of everything
            # to make sure we dont get a messed up system state with wrong list/diff from
            # dynamically switching cache on and off.
            self.connection_cache = None

        # TODO Instead of one or the other, we should have "two layer behaviors" with different frequencies
        # Fast loop checking only diff
        # Slow loop checking full state
        # It will allow recovering from any mistakes because of wrong diffs (update speed/race conditions/etc.)
        if self.enable_cache:
            if self.connection_cache is None:  # Building Connection Cache Proxy if needed
                self.connection_cache = connection_cache_proxy_create(self._proxy_cb)
                if self.connection_cache:
                    self.enable_cache = True
                else:
                    self.enable_cache = False

        # TMP until it s implemented in the connection cache
        # Because the cache doesnt currently do it
        params = set(rospy.get_param_names())
        # determining params diff despite lack of API
        params_dt = DiffTuple(
            added=[p for p in params if p not in self.params_available],
            removed=[p for p in self.params_available if p not in params]
        )

        # If we have the connection_cache and a callback setup we process the diff (and maybe param changes)
        if self.connection_cache and (params_dt.added or params_dt.removed or backedup_complete_cb_ss is not None or self.cb_ss.qsize() > 0):
            try:
                if backedup_complete_cb_ss is not None:
                    # using the message backed up after latest diff process
                    cb_ss = backedup_complete_cb_ss
                    backedup_complete_cb_ss = None  #CAREFUL : untested...
                else:
                    cb_ss = self.cb_ss.get_nowait()
                # print("CC MSG !")  # if we didn't except on empty queue, so we got a message

            except Queue.Empty:

                return self.update_nodelta(params_dt)

            else:
                # if there was no change,
                # it means it s the first and we need to initialize with the full list
                if cb_ss.added is None and cb_ss.removed is None:
                    #print("CC COMPLETE !")
                    publishers = cb_ss.complete.get('publishers', [])
                    subscribers = cb_ss.complete.get('subscribers', [])
                    services = cb_ss.complete.get('services', [])
                    # NOT YET...
                    #params = cb_ss.complete.get('params', params)
                    topic_types = cb_ss.complete.get('topic_types', [])
                    service_types = cb_ss.complete.get('service_types', [])

                    # To convert to ROS master API format :
                    publishers = [[p, [n[0] for n in nset]] for p, nset in publishers.iteritems()]
                    subscribers = [[s, [n[0] for n in nset]] for s, nset in subscribers.iteritems()]
                    services = [[s, [n[0] for n in nset]] for s, nset in services.iteritems()]

                    #print("CC COMPLETEDDDD !")

                    # we go back to normal flow
                    #print("UPDATE FULLSTATE ON CACHE LIST")
                    return self.update_fullstate(publishers, subscribers, services, params, topic_types, service_types)

                else:  # we have a delta, we can use it directly and skip the rest

                    #print("CC DELTA !")
                    next_cb_ss = cb_ss
                    while self.cb_ss.qsize() > 0 and not (next_cb_ss.added is None and next_cb_ss is None):
                        # merging multiple diff messages as fast as possible, until the next complete status
                        next_cb_ss = self.cb_ss.get_nowait()
                        if not (next_cb_ss.added is None and next_cb_ss is None):
                            cb_ss = CacheTuple(
                                complete=None,  # this is ignored here
                                # CAREFUL one of added or remove can still be None here...
                                added=connection_cache_merge_marshalled(next_cb_ss.added or {}, cb_ss.added),
                                removed=connection_cache_merge_marshalled(next_cb_ss.removed or {}, cb_ss.removed)
                            )
                        else:  # we need to pass next_cb_ss to handle complete list again...
                            backedup_complete_cb_ss = next_cb_ss

                    added_publishers = cb_ss.added.get('publishers', [])
                    added_subscribers = cb_ss.added.get('subscribers', [])
                    added_services = cb_ss.added.get('services', [])
                    # NOT YET
                    #added_params = cb_ss.added.get('params', params_dt.added)
                    added_params = params_dt.added
                    added_topic_types = cb_ss.added.get('topic_types', [])
                    added_service_types = cb_ss.added.get('service_types', [])

                    removed_publishers = cb_ss.removed.get('publishers', [])
                    removed_subscribers = cb_ss.removed.get('subscribers', [])
                    removed_services = cb_ss.removed.get('services', [])
                    # NOT YET
                    #removed_params = cb_ss.removed.get('params', params_dt.removed)
                    removed_params = params_dt.removed
                    removed_topic_types = cb_ss.removed.get('topic_types', [])
                    removed_service_types = cb_ss.removed.get('service_types', [])

                    #print("UPDATE STATEDELTA ON CACHE DIFF")
                    return self.update_statedelta(
                        added_publishers, added_subscribers, added_services, added_params, added_topic_types, added_service_types,
                        removed_publishers, removed_subscribers, removed_services, removed_params, removed_topic_types, removed_service_types
                    )
        elif not self.connection_cache:  # make sure we are not using connection cache (otherwise state representations might be out of sync !!)
            #print("GETTING STATE FROM MASTER")
            publishers, subscribers, services, params, topic_types, service_types = self.retrieve_system_state()  # This will call the master

            #print("UPDATE FULLSTATE")
            # NOTE : we want to be certain here that we do not mix full state representation from master with representation from cache (out of sync !!!)
            return self.update_fullstate(publishers, subscribers, services, params, topic_types, service_types)
        else:
            # This happens when the update is triggered, we are supposed to use the cache, but we got no message from the connection cache proxy.
            # WE DO NOT WANT TO GET FULLSTATE FROM CACHE, since it is out of date (ex: removed publishers are not removed until next message...)
            # print("GETTING STATE FROM CACHE")
            # publishers, subscribers, services, params, topic_types, service_types = self.retrieve_system_state()  # This will call the cache
            #
            # print("UPDATE FULLSTATE")
            # return self.update_fullstate(publishers, subscribers, services, params, topic_types, service_types)
            # ==> We need to wait for next message...
            return self.update_nodelta(params_dt)

    def update_fullstate(self, publishers, subscribers, services, params, topic_types, service_types):
        # NORMAL full update
        self._debug_logger.debug("""SYSTEM STATE :
                    - publishers : {publishers}
                    - subscribers : {subscribers}
                    - services : {services}
                    - params : {params}
                    - topic_types : {topic_types}
                    - service_types : {service_types}
                """.format(**locals()))

        # TODO : unify with the reset behavior in case of cache...

        # Needs to be done first, since topic algorithm depends on it
        # print("PARAMS : {params}".format(**locals()))
        params_if_dt = self.params_if_pool.update(params=params)
        # print("PARAM IF DT : {params_if_dt}".format(**locals()))

        # print("SERVICES : {services}".format(**locals()))
        services_if_dt = self.services_if_pool.update(services, service_types)
        # print("SERVICE IF DT : {services_if_dt}".format(**locals()))

        # print("SUBSCRIBERS : {subscribers}".format(**locals()))
        subscribers_if_dt = self.subscribers_if_pool.update(subscribers, topic_types)
        # print("SUBSCRIBER IF DT : {subscribers_if_dt}".format(**locals()))

        # print("PUBLISHERS : {publishers}".format(**locals()))
        publishers_if_dt = self.publishers_if_pool.update(publishers, topic_types)
        # print("PUBLISHER IF DT : {publishers_if_dt}".format(**locals()))

        dt = DiffTuple(
            added=params_if_dt.added + services_if_dt.added + subscribers_if_dt.added + publishers_if_dt.added,
            removed=params_if_dt.removed + services_if_dt.removed + subscribers_if_dt.removed + publishers_if_dt.removed
        )

        self._debug_logger.debug("""
                    ROS INTERFACE ADDED : {dt.added}
                    ROS INTERFACE REMOVED : {dt.removed}
                """.format(**locals()))

        return dt

    def update_nodelta(self, params_dt):
        """
        Running updates of nothing (workaround until params are handled by cache)
        :param params_dt:
        :return:
        """
        # we re NOT done here, we might still need to update params
        params_if_dt = self.params_if_pool.update_delta(params_dt=params_dt)

        topic_types_dt = DiffTuple(
            added=[],
            removed=[]  # shouldnt matter
        )

        services_if_dt = DiffTuple([], [])
        # TMP : NOT dropping topics early (just be patient and wait for the cache callback to come...)
        # topics_if_dt = self.topics_pool.update_delta(topics_dt, topic_types_dt)
        subscribers_if_dt = DiffTuple([], [])
        publishers_if_dt = DiffTuple([], [])

        # and here we need to return to not do the normal full update
        dt = DiffTuple(
            added=params_if_dt.added + services_if_dt.added + subscribers_if_dt.added + publishers_if_dt.added,
            removed=params_if_dt.removed + services_if_dt.removed + subscribers_if_dt.removed + publishers_if_dt.removed
        )

        self._debug_logger.debug("""
                            ROS INTERFACE DIFF ADDED : {dt.added}
                            ROS INTERFACE DIFF REMOVED : {dt.removed}
                        """.format(**locals()))

        return dt

    def update_statedelta(
            self,
            added_publishers, added_subscribers, added_services, added_params, added_topic_types, added_service_types,
            removed_publishers, removed_subscribers, removed_services, removed_params, removed_topic_types, removed_service_types
    ):
        params_dt = DiffTuple(
            added=added_params,
            removed=removed_params
        )

        # Needs to be done first, since topic algorithm depends on it
        params_if_dt = self.params_if_pool.update_delta(params_dt=params_dt)

        # here we need to get only the nodes' names to match ROs master API format
        services_dt = DiffTuple(
            added=[[k, [n[0] for n in nset]] for k, nset in added_services.iteritems()],
            removed=[[k, [n[0] for n in nset]] for k, nset in removed_services.iteritems()]
        )

        service_types_dt = DiffTuple(
            added=added_service_types,
            removed=removed_service_types
        )

        services_if_dt = self.services_if_pool.update_delta(services_dt, service_types_dt)

        # here we need to get only the nodes' names to match ROs master API format
        # CAREFUL about unicity here, a Pub|Sub can have multiple endpoint within the same node...
        # We do not care about that information. If we have to care, then we should pass also the uri...
        publishers_dt = DiffTuple(
            added=[[k, [n[0] for n in set(nset)]] for k, nset in added_publishers.iteritems()],
            removed=[[k, [n[0] for n in set(nset)]] for k, nset in removed_publishers.iteritems()]
        )
        subscribers_dt = DiffTuple(
            added=[[k, [n[0] for n in set(nset)]] for k, nset in added_subscribers.iteritems()],
            removed=[[k, [n[0] for n in set(nset)]] for k, nset in removed_subscribers.iteritems()]
        )

        # NOW DONE IN update_delta
        # # CAREFUL, topic interface by itself also makes the topic detected on system
        # # Check if there are any pyros interface with it and ignore them
        # topics_to_drop = TopicBack.get_interface_only_topics()
        # # TODO : simplify : same as early_topics_to_drop ?
        #
        # # filtering the topic dict
        # for td, tnode in added_topics.iteritems():
        #     tnode = [n for n in tnode if n not in topics_to_drop.get(td, [])]
        #
        # added_topics_list = [[td, added_topics[td]] for td in added_topics if added_topics[td]]  # filtering out topics with no endpoints
        #
        # # we also need to simulate topic removal here (only names), to trigger a cleanup of interface if it s last one
        # removed_topics_list = [[td, removed_topics[td]] for td in removed_topics] + early_topics_to_drop

        topic_types_dt = DiffTuple(
            added=added_topic_types,
            removed=removed_topic_types
        )

        subscribers_if_dt = self.subscribers_if_pool.update_delta(subscribers_dt, topic_types_dt)
        publishers_if_dt = self.publishers_if_pool.update_delta(publishers_dt, topic_types_dt)

        if publishers_if_dt.added or publishers_if_dt.removed:
            self._debug_logger.debug(
                rospy.get_name() + " Pyros.rosinterface : Publishers Delta {publishers_if_dt}".format(
                    **locals()))
        if subscribers_if_dt.added or subscribers_if_dt.removed:
            self._debug_logger.debug(
                rospy.get_name() + " Pyros.rosinterface : Subscribers Delta {subscribers_if_dt}".format(
                    **locals()))
        if services_if_dt.added or services_if_dt.removed:
            self._debug_logger.debug(
                rospy.get_name() + " Pyros.rosinterface : Services Delta {services_if_dt}".format(
                    **locals()))

        # TODO : put that in debug log and show based on python logger configuration
        # print("Pyros ROS interface UPDATE")
        # print("Params ADDED : {0}".format([p for p in params_dt.added]))
        # print("Params GONE : {0}".format([p for p in params_dt.removed]))
        # print("Topics ADDED : {0}".format([t[0] for t in topics_dt.added] + early_topics_dt.added))
        # print("Topics GONE : {0}".format([t[0] for t in topics_dt.removed] + early_topics_dt.removed))
        # print("Srvs ADDED: {0}".format([s[0] for s in services_dt.added]))
        # print("Srvs GONE: {0}".format([s[0] for s in services_dt.removed]))

        # update_on_diff wants only names
        # dt = super(RosInterface, self).update_on_diff(
        #         DiffTuple([s[0] for s in services_dt.added], [s[0] for s in services_dt.removed]),
        #         DiffTuple([t[0] for t in topics_dt.added] + early_topics_dt.added, [t[0] for t in topics_dt.removed] + early_topics_dt.removed),
        #         # Careful params_dt has a different content than service and topics, due to different ROS API
        #         # TODO : make this as uniform as possible
        #         DiffTuple([p for p in params_dt.added], [p for p in params_dt.removed])
        # )

        # and here we need to return to not do the normal full update
        dt = DiffTuple(
            added=params_if_dt.added + services_if_dt.added + subscribers_if_dt.added + publishers_if_dt.added,
            removed=params_if_dt.removed + services_if_dt.removed + subscribers_if_dt.removed + publishers_if_dt.removed
        )

        self._debug_logger.debug("""
                                        ROS INTERFACE DIFF ADDED : {dt.added}
                                        ROS INTERFACE DIFF REMOVED : {dt.removed}
                                    """.format(**locals()))

        return dt


    def _proxy_cb(self, system_state, added_system_state, lost_system_state):
        self.cb_ss.put(CacheTuple(
            complete=connection_cache_marshall(system_state),
            added=connection_cache_marshall(added_system_state) if added_system_state is not None else None,
            removed=connection_cache_marshall(lost_system_state) if lost_system_state is not None else None
        ))






#!/usr/bin/env python
# Software License Agreement (BSD License)
#
# Copyright (c) 2012, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import roslib
import rospy

from threading import Lock

""" ros_loader contains methods for dynamically loading ROS message classes at
runtime.  It's achieved by using roslib to load the manifest files for the
package that the respective class is contained in.

Methods typically return the requested class or instance, or None if not found
"""

# Variable containing the loaded classes
_loaded_msgs = {}
_loaded_srvs = {}
_msgs_lock = Lock()
_srvs_lock = Lock()


class InvalidTypeStringException(Exception):
    def __init__(self, typestring):
        Exception.__init__(self, "{0!s} is not a valid type string".format(typestring))


class InvalidPackageException(Exception):
    def __init__(self, package, original_exception):
        Exception.__init__(self,
           "Unable to load the manifest for package {0!s}. Caused by: {1!s}".format(package, original_exception.message)
       )


class InvalidModuleException(Exception):
    def __init__(self, modname, subname, original_exception):
        Exception.__init__(self,
           "Unable to import {0!s}.{1!s} from package {2!s}. Caused by: {3!s}".format(modname, subname, modname, str(original_exception))
        )


class InvalidClassException(Exception):
    def __init__(self, modname, subname, classname, original_exception):
        Exception.__init__(self,
           "Unable to import {0!s} class {1!s} from package {2!s}. Caused by {3!s}".format(subname, classname, modname, str(original_exception))
        )


def get_message_class(typestring):
    """ Loads the message type specified.

    Returns the loaded class, or throws exceptions on failure """
    return _get_msg_class(typestring)


def get_service_class(typestring):
    """ Loads the service type specified.

    Returns the loaded class, or None on failure """
    return _get_srv_class(typestring)


def get_message_instance(typestring):
    """ If not loaded, loads the specified type.
    Then returns an instance of it, or None. """
    cls = get_message_class(typestring)
    return cls()


def get_service_instance(typestring):
    """ If not loaded, loads the specified type.
    Then returns an instance of it, or None. """
    cls = get_service_class(typestring)
    return cls()


def get_service_request_instance(typestring):
    cls = get_service_class(typestring)
    return cls._request_class()


def get_service_response_instance(typestring):
    cls = get_service_class(typestring)
    return cls._response_class()


def _get_msg_class(typestring):
    """ If not loaded, loads the specified msg class then returns an instance
    of it

    Throws various exceptions if loading the msg class fails """
    global _loaded_msgs, _msgs_lock
    return _get_class(typestring, "msg", _loaded_msgs, _msgs_lock)


def _get_srv_class(typestring):
    """ If not loaded, loads the specified srv class then returns an instance
    of it

    Throws various exceptions if loading the srv class fails """
    global _loaded_srvs, _srvs_lock
    return _get_class(typestring, "srv", _loaded_srvs, _srvs_lock)


def _get_class(typestring, subname, cache, lock):
    """ If not loaded, loads the specified class then returns an instance
    of it.

    Loaded classes are cached in the provided cache dict

    Throws various exceptions if loading the msg class fails """

    # First, see if we have this type string cached
    cls = _get_from_cache(cache, lock, typestring)
    if cls is not None:
        return cls

    # Now normalise the typestring
    modname, classname = _splittype(typestring)
    norm_typestring = modname + "/" + classname

    # Check to see if the normalised type string is cached
    cls = _get_from_cache(cache, lock, norm_typestring)
    if cls is not None:
        return cls

    # Load the class
    cls = _load_class(modname, subname, classname)

    # Cache the class for both the regular and normalised typestring
    _add_to_cache(cache, lock, typestring, cls)
    _add_to_cache(cache, lock, norm_typestring, cls)

    return cls


def _load_class(modname, subname, classname):
    """ Loads the manifest and imports the module that contains the specified
    type.

    Logic is similar to that of roslib.message.get_message_class, but we want
    more expressive exceptions.

    Returns the loaded module, or None on failure """
    global loaded_modules

    try:
        # roslib maintains a cache of loaded manifests, so no need to duplicate
        roslib.launcher.load_manifest(modname)
    except Exception as exc:
        raise InvalidPackageException(modname, exc)

    try:
        pypkg = __import__('{0!s}.{1!s}'.format(modname, subname))
    except Exception as exc:
        raise InvalidModuleException(modname, subname, exc)

    try:
        return getattr(getattr(pypkg, subname), classname)
    except Exception as exc:
        raise InvalidClassException(modname, subname, classname, exc)


def _splittype(typestring):
    """ Split the string the / delimiter and strip out empty strings

    Performs similar logic to roslib.names.package_resource_name but is a bit
    more forgiving about excess slashes
    """
    splits = [x for x in typestring.split("/") if x]
    if len(splits) == 2:
        return splits
    raise InvalidTypeStringException(typestring)


def _add_to_cache(cache, lock, key, value):
    lock.acquire()
    cache[key] = value
    lock.release()


def _get_from_cache(cache, lock, key):
    """ Returns the value for the specified key from the cache.
    Locks the lock before doing anything. Returns None if key not in cache """
    lock.acquire()
    ret = None
    if key in cache:
        ret = cache[key]
    lock.release()
    return ret


from __future__ import absolute_import

from . import deffile
from .util import get_json_bool, type_str, load_type

def get_all_msg_types(msg, skip_this=False, type_set=None):
    if type_set is None:
        type_set = set()
    if not skip_this:
        if msg in type_set:
            return type_set
        type_set.add(msg)
    for slot_type in msg._slot_types:
        if '/' not in slot_type:
            continue
        type_set = get_all_msg_types(load_type(slot_type), type_set=type_set)
    return type_set


def get_msg_definitions(msg, skip_this=False):
    type_set = get_all_msg_types(msg, skip_this=skip_this)

    msg_dfns = []
    for msg_type in type_set:
        dfn = deffile.ROSStyleDefinition('msg', type_str(msg_type), ['msg'])
        for field_name, field_type in zip(msg_type.__slots__, msg_type._slot_types):
            dfn.segment(0).append((field_name, field_type))
        msg_dfns.append(dfn)
    return msg_dfns


def get_definitions(services=None, topics=None, actions=None):
    if services is None:
        services = []
    if topics is None:
        topics = []
    if actions is None:
        actions = []
    service_dfns = []
    action_dfns = []
    msg_dfns = []

    type_set = set()
    for service in services:
        dfn = deffile.ROSStyleDefinition('srv', service.get('rostype_name'), ['request', 'response'])
        service_type = load_type(service.get('rostype_name'))
        for field_name, field_type in service.get('srvtype',[None, None])[0].iteritems():
            dfn.segment(0).append((field_name, field_type))
            type_set = get_all_msg_types(service_type._request_class, skip_this=True, type_set=type_set)
        for field_name, field_type in service.get('srvtype',[None, None])[1].iteritems():
            dfn.segment(1).append((field_name, field_type))
            type_set = get_all_msg_types(service_type._response_class, skip_this=True, type_set=type_set)
        service_dfns.append(dfn)

    for action in actions:
        dfn = deffile.ROSStyleDefinition('action', action.rostype_name, ['goal', 'result', 'feedback'])
        for field_name, field_type in zip(action.rostype_goal.__slots__, action.rostype_goal._slot_types):
            dfn.segment(0).append((field_name, field_type))
            type_set = get_all_msg_types(action.rostype_goal, skip_this=True, type_set=type_set)
        for field_name, field_type in zip(action.rostype_result.__slots__, action.rostype_result._slot_types):
            dfn.segment(1).append((field_name, field_type))
            type_set = get_all_msg_types(action.rostype_result, skip_this=True, type_set=type_set)
        for field_name, field_type in zip(action.rostype_feedback.__slots__, action.rostype_feedback._slot_types):
            dfn.segment(2).append((field_name, field_type))
            type_set = get_all_msg_types(action.rostype_feedback, skip_this=True, type_set=type_set)
        action_dfns.append(dfn)

    for topic in topics:
        topic_type = load_type(topic.get('rostype_name'))
        type_set = get_all_msg_types(topic_type, type_set=type_set)

    for msg_type in type_set:
        dfn = deffile.ROSStyleDefinition('msg', type_str(msg_type), ['msg'])
        for field_name, field_type in zip(msg_type.__slots__, msg_type._slot_types):
            dfn.segment(0).append((field_name, field_type))
        msg_dfns.append(dfn)

    return msg_dfns + service_dfns + action_dfns


def manifest(services, topics, actions, full=False):
    dfile = deffile.DefFile()
    dfile.manifest.def_type = 'Node'

    if services:
        service_section = deffile.INISection('Services')
        for service_name, service in services.iteritems():
            service_section.fields[service_name] = service.rostype_name
        dfile.add_section(service_section)

    if actions:
        action_section = deffile.INISection('Actions')
        for action_name, action in actions.iteritems():
            action_section.fields[action_name] = action.rostype_name
        dfile.add_section(action_section)

    topics_section = deffile.INISection('Topics')
    publish_section = deffile.INISection('Publishes')
    subscribe_section = deffile.INISection('Subscribes')

    for topic_name, topic in topics.iteritems():
        if topic.allow_sub:
            publish_section.fields[topic_name] = topic.rostype_name
        if topic.allow_pub:
            subscribe_section.fields[topic_name] = topic.rostype_name

    if topics_section.fields:
        dfile.add_section(topics_section)
    if publish_section.fields:
        dfile.add_section(publish_section)
    if subscribe_section.fields:
        dfile.add_section(subscribe_section)

    if full:
        dfns = get_definitions(services=services.itervalues(), topics=topics.itervalues(), actions=actions.itervalues())
        [dfile.add_definition(dfn) for dfn in dfns]

    return dfile


def describe_service(service_name, service, full=False):
    dfile = deffile.DefFile()
    dfile.manifest.def_type = 'Service'
    dfile.manifest['Name'] = service_name
    dfile.manifest['Type'] = service.get('rostype_name', 'Unknown')

    if full:
        dfns = get_definitions(services=[service])
        [dfile.add_definition(dfn) for dfn in dfns]

    return dfile


# Interestingly this is used from rostful, to make sense of the data returned by pyros client, not from pyros itself...
# TODO : maybe need to move it ?
# TODO : check, maybe same with some other methods here...
def describe_topic(topic_name, topic, full=False):
    dfile = deffile.DefFile()
    dfile.manifest.def_type = 'Topic'
    dfile.manifest['Name'] = topic_name
    dfile.manifest['Type'] = topic.get('rostype_name', 'Unknown')
    # this is obsolete, now each Topic instance does both...
    #dfile.manifest['Publishes'] = get_json_bool(topic.allow_sub)
    #dfile.manifest['Subscribes'] = get_json_bool(topic.allow_pub)

    if full:
        dfns = get_definitions(topics=[topic])
        [dfile.add_definition(dfn) for dfn in dfns]

    return dfile




from __future__ import absolute_import

from .mockservicepool import MockServicePool
from .mocksubscriberpool import MockSubscriberPool
from .mockpublisherpool import MockPublisherPool
from .mockparampool import MockParamPool

from ..baseinterface import BaseInterface


class MockInterface(BaseInterface):

    """
    MockInterface. A simple mock following ROS design.
    Userful to test the multiprocess parts of pyros without worrying about real connexion with ROS.
    """
    def __init__(self, publishers=None, subscribers=None, services=None, params=None):

        publishers_pool = MockPublisherPool(publishers)
        subscribers_pool = MockSubscriberPool(subscribers)
        services_pool = MockServicePool(services)
        params_pool = MockParamPool(params)

        super(MockInterface, self).__init__(publishers_pool, subscribers_pool, services_pool, params_pool)

    #
    # # REQUESTED
    # @property
    # def services_args(self):
    #     return self.service_pool.transients_args
    #
    # @property
    # def topics_args(self):
    #     return self.topic_pool.transients_args
    #
    # @property
    # def params_args(self):
    #     return self.param_pool.transients_args
    #
    # # AVAILABLE
    # @property
    # def services_available(self):
    #     return self.service_pool.available
    #
    # @property
    # def topics_available(self):
    #     return self.topic_pool.available
    #
    # @property
    # def params_available(self):
    #     return self.param_pool.available
    #
    # # INTERFACED
    # @property
    # def services(self):
    #     return self.service_pool.transients
    #
    # @property
    # def topics(self):
    #     return self.topic_pool.transients
    #
    # @property
    # def params(self):
    #     return self.param_pool.transients
    #
    # # EXPOSE
    # def expose_services(self, svc_regex):
    #     return self.service_pool.expose_transients_regex(svc_regex)
    #
    # def expose_topics(self, tpc_regex):
    #     return self.topic_pool.expose_transients_regex(tpc_regex)
    #
    # def expose_params(self, prm_regex):
    #     return self.param_pool.expose_transients_regex(prm_regex)
    #
    #MOCK
    def mock_publisher(self, tpc_name, tpc_type):
        return self.publishers_if_pool.mock(tpc_name, tpc_type)

    def mock_subscriber(self, tpc_name, tpc_type):
        return self.subscribers_if_pool.mock(tpc_name, tpc_type)

    def mock_service(self, svc_name, svc_type):
        return self.services_if_pool.mock(svc_name, svc_type)

    def mock_param(self, prm_name, prm_type):
        return self.params_if_pool.mock(prm_name, prm_type)
    #
    # #CHANGE DIFF
    # def services_change_diff(self, appeared, gone):
    #     return self.service_pool.transient_change_diff(appeared, gone)
    #
    # def topics_change_diff(self, appeared, gone):
    #     return self.topic_pool.transient_change_diff(appeared, gone)
    #
    # def params_change_diff(self, appeared, gone):
    #     return self.param_pool.transient_change_diff(appeared, gone)
    #
    # #CHAGNE DETECT
    # def services_change_detect(self):
    #     return self.service_pool.transient_change_detect()
    #
    # def topics_change_detect(self):
    #     return self.topic_pool.transient_change_detect()
    #
    # def param_change_detect(self):
    #     return self.param_pool.transient_change_detect()
    #
    # #UPDATE
    # def update_services(self, add_names, remove_names):
    #     return self.service_pool.update_transients(add_names, remove_names)
    #
    # def update_topics(self, add_names, remove_names):
    #     return self.topic_pool.update_transients(add_names, remove_names)
    #
    # def update_params(self, add_names, remove_names):
    #     return self.param_pool.update_transients(add_names, remove_names)

    def update(self):
        self.publishers_if_pool.update()
        self.subscribers_if_pool.update()
        self.services_if_pool.update()
        self.params_if_pool.update()



#!/usr/bin/env python
# Software License Agreement (BSD License)
#
# Copyright (c) 2012, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from .api import roslib_safe as roslib
from .api import rospy_safe as rospy
# TODO : if possible use rospy internals

from ..exceptions import PyrosException

import re
import string
from base64 import standard_b64encode, standard_b64decode

# Utils function to get message type from ros or its dict representation
#outputs message structure as string (useful ?)
def get_msg(msg):
    return '\n'.join(['{0!s} {1!s}'.format(*line) for line in zip(msg._slot_types, msg.__slots__)])

#outputs message structure as dict
def get_msg_dict(msg):
    return dict(zip(msg.__slots__,msg._slot_types))

type_map = {
   "bool":    ["bool"],
   "int":     ["int8", "byte", "uint8", "char",
               "int16", "uint16", "int32", "uint32",
               "int64", "uint64", "float32", "float64"],
   "float":   ["float32", "float64"],
   "str":     ["string"],
   "unicode": ["string"],
   "long":    ["uint64"]
}
primitive_types = [bool, int, long, float]
string_types = [str, unicode]
list_types = [list, tuple]
ros_time_types = ["time", "duration"]
ros_primitive_types = ["bool", "byte", "char", "int8", "uint8", "int16",
                       "uint16", "int32", "uint32", "int64", "uint64",
                       "float32", "float64", "string"]
ros_header_types = ["Header", "std_msgs/Header", "roslib/Header"]
ros_binary_types = ["uint8[]", "char[]"]
list_braces = re.compile(r'\[[^\]]*\]')


# CAREFUL : exceptions must be pickleable ( we need to pass all arguments to the superclass )
class InvalidMessageException(PyrosException):
    def __init__(self, inst):
        super(InvalidMessageException, self).__init__(inst)
        self.inst = inst
        self.excmsg = "Unable to extract message values from {0!s} instance".format(type(inst).__name__)

    @property
    def message(self):
        return self.excmsg

PyrosException.register(InvalidMessageException)


class NonexistentFieldException(PyrosException):
    def __init__(self, basetype, fields):
        super(NonexistentFieldException, self).__init__(basetype, fields)
        self.basetype = basetype
        self.fields = fields
        self.excmsg = "Message type {0!s} does not have a field {1!s}".format(basetype, '.'.join(fields))

    @property
    def message(self):
        return self.excmsg

PyrosException.register(NonexistentFieldException)


class FieldTypeMismatchException(Exception):
    def __init__(self, roottype, fields, expected_type, found_type):
        super(FieldTypeMismatchException, self).__init__(roottype, fields, expected_type, found_type)
        self.roottype = roottype
        self.fields = fields
        self.expected_type = expected_type
        self.found_type = found_type

        if self.roottype == self.expected_type:
            self.excmsg = "Expected a JSON object for type {0!s} but received a {1!s}".format(self.roottype, self.found_type)
        else:
            self.excmsg = "{0!s} message requires a {1!s} for field {2!s}, but got a {3!s}".format(self.roottype, self.expected_type, '.'.join(fields), self.found_type)

    @property
    def message(self):
        return self.excmsg

PyrosException.register(FieldTypeMismatchException)



### This is a JSON -> ROS conversion module

def extract_values(inst):
    rostype = getattr(inst, "_type", None)
    if rostype is None:
        raise InvalidMessageException()
    return _from_inst(inst, rostype)
dump = extract_values

def populate_instance(msg, inst):
    """ Returns an instance of the provided class, with its fields populated
    according to the values in msg """
    # if we need to populate an Empty message, we return it already
    return inst if inst is None else _to_inst(msg, inst._type, inst._type, inst)
#load = populate_instance

def _from_inst(inst, rostype):
    # Special case for uint8[], we base64 encode the string
    if rostype in ros_binary_types:
        return standard_b64encode(inst)

    # Check for time or duration
    if rostype in ros_time_types:
        return {"secs": inst.secs, "nsecs": inst.nsecs}

    # Check for primitive types
    if rostype in ros_primitive_types:
        return inst

    # Check if it's a list or tuple
    if type(inst) in list_types:
        return _from_list_inst(inst, rostype)

    # Assume it's otherwise a full ros msg object
    return _from_object_inst(inst, rostype)


def _from_list_inst(inst, rostype):
    # Can duck out early if the list is empty
    if not inst:
        return []

    # Remove the list indicators from the rostype
    rostype = list_braces.sub("", rostype)
    
    # Shortcut for primitives
    if rostype in ros_primitive_types:
        return list(inst)

    # Call to _to_inst for every element of the list
    return [_from_inst(x, rostype) for x in inst]


def _from_object_inst(inst, rostype):
    # Create an empty dict then populate with values from the inst
    msg = {}
    for field_name, field_rostype in zip(inst.__slots__, inst._slot_types):
        field_inst = getattr(inst, field_name)
        msg[field_name] = _from_inst(field_inst, field_rostype)
    return msg


def _to_inst(msg, rostype, roottype, inst=None, stack=None):
    # Check if it's uint8[], and if it's a string, try to b64decode
    if stack is None:
        stack = []
    if rostype in ros_binary_types:
        return _to_binary_inst(msg)

    # Check the type for time or rostime
    if rostype in ros_time_types:
        return _to_time_inst(msg, rostype, inst)

    # Check to see whether this is a primitive type
    if rostype in ros_primitive_types:
        return _to_primitive_inst(msg, rostype, roottype, stack)

    # Check whether we're dealing with a list type
    if inst is not None and type(inst) in list_types:
        return _to_list_inst(msg, rostype, roottype, inst, stack)

    # Otherwise, the type has to be a full ros msg type, so msg must be a dict
    if inst is None:
        inst = _get_msg_class(rostype)()

    return _to_object_inst(msg, rostype, roottype, inst, stack)


def _to_binary_inst(msg):
    if type(msg) in string_types:
        try:
            return standard_b64decode(msg)
        except:
            return msg
    else:
        try:
            return str(bytearray(msg))
        except:
            return msg


def _to_time_inst(msg, rostype, inst=None):
    # Create an instance if we haven't been provided with one
    if rostype == "time" and msg == "now":
        return rospy.get_rostime()

    if inst is None:
        if rostype == "time":
            inst = rospy.rostime.Time()
        elif rostype == "duration":
            inst = rospy.rostime.Duration()
        else:
            return None

    # Copy across the fields
    for field in ["secs", "nsecs"]:
        try:
            if field in msg:
                setattr(inst, field, msg[field])
        except TypeError:
            continue

    return inst


def _to_primitive_inst(msg, rostype, roottype, stack):
    # Typecheck the msg
    msgtype = type(msg)
    if msgtype in primitive_types and rostype in type_map[msgtype.__name__]:
        return msg
    elif msgtype in string_types and rostype in type_map[msgtype.__name__]:
        return msg.encode("ascii", "ignore")
    raise FieldTypeMismatchException(roottype, stack, rostype, msgtype)


def _to_list_inst(msg, rostype, roottype, inst, stack):
    # Typecheck the msg
    if type(msg) not in list_types:
        raise FieldTypeMismatchException(roottype, stack, rostype, type(msg))

    # Can duck out early if the list is empty
    if len(msg) == 0:
        return []

    # Remove the list indicators from the rostype
    rostype = list_braces.sub("", rostype)

    # Call to _to_inst for every element of the list
    return [_to_inst(x, rostype, roottype, None, stack) for x in msg]


def _to_object_inst(msg, rostype, roottype, inst, stack):
    # Typecheck the msg
    if type(msg) is not dict:
        raise FieldTypeMismatchException(roottype, stack, rostype, type(msg))

    # Substitute the correct time if we're an std_msgs/Header
    if rostype in ros_header_types:
        inst.stamp = rospy.get_rostime()

    inst_fields = dict(zip(inst.__slots__, inst._slot_types))

    for field_name in msg:
        # Add this field to the field stack
        field_stack = stack + [field_name]

        # Raise an exception if the msg contains a bad field
        if not field_name in inst_fields:
            raise NonexistentFieldException(roottype, field_stack)

        field_rostype = inst_fields[field_name]
        field_inst = getattr(inst, field_name)

        field_value = _to_inst(msg[field_name], field_rostype,
                    roottype, field_inst, field_stack)

        setattr(inst, field_name, field_value)

    return inst

from threading import Lock
# Variable containing the loaded classes
_loaded_msgs = {}
_loaded_srvs = {}
_msgs_lock = Lock()
_srvs_lock = Lock()


class InvalidTypeStringException(Exception):
    def __init__(self, typestring):
        Exception.__init__(self, "{0!s} is not a valid type string".format(typestring))


class InvalidPackageException(Exception):
    def __init__(self, package, original_exception):
        Exception.__init__(self,
           "Unable to load the manifest for package {0!s}. Caused by: {1!s}".format(package, original_exception.message)
       )


class InvalidModuleException(Exception):
    def __init__(self, modname, subname, original_exception):
        Exception.__init__(self,
           "Unable to import {0!s}.{1!s} from package {2!s}. Caused by: {3!s}".format(modname, subname, modname, str(original_exception))
        )


class InvalidClassException(Exception):
    def __init__(self, modname, subname, classname, original_exception):
        Exception.__init__(self,
           "Unable to import {0!s} class {1!s} from package {2!s}. Caused by {3!s}".format(subname, classname, modname, str(original_exception))
        )


def _get_msg_class(typestring):
    """ If not loaded, loads the specified msg class then returns an instance
    of it

    Throws various exceptions if loading the msg class fails """
    global _loaded_msgs, _msgs_lock
    return _get_class(typestring, "msg", _loaded_msgs, _msgs_lock)


def _get_srv_class(typestring):
    """ If not loaded, loads the specified srv class then returns an instance
    of it

    Throws various exceptions if loading the srv class fails """
    global _loaded_srvs, _srvs_lock
    return _get_class(typestring, "srv", _loaded_srvs, _srvs_lock)


def _get_class(typestring, subname, cache, lock):
    """ If not loaded, loads the specified class then returns an instance
    of it.

    Loaded classes are cached in the provided cache dict

    Throws various exceptions if loading the msg class fails """

    # First, see if we have this type string cached
    cls = _get_from_cache(cache, lock, typestring)
    if cls is not None:
        return cls

    # Now normalise the typestring
    modname, classname = _splittype(typestring)
    norm_typestring = modname + "/" + classname

    # Check to see if the normalised type string is cached
    cls = _get_from_cache(cache, lock, norm_typestring)
    if cls is not None:
        return cls

    # Load the class
    cls = _load_class(modname, subname, classname)

    # Cache the class for both the regular and normalised typestring
    _add_to_cache(cache, lock, typestring, cls)
    _add_to_cache(cache, lock, norm_typestring, cls)

    return cls


def _load_class(modname, subname, classname):
    """ Loads the manifest and imports the module that contains the specified
    type.

    Logic is similar to that of roslib.message.get_message_class, but we want
    more expressive exceptions.

    Returns the loaded module, or None on failure """
    global loaded_modules

    try:
        # roslib maintains a cache of loaded manifests, so no need to duplicate
        roslib.launcher.load_manifest(modname)
    except Exception as exc:
        raise InvalidPackageException(modname, exc)

    try:
        pypkg = __import__('{0!s}.{1!s}'.format(modname, subname))
    except Exception as exc:
        raise InvalidModuleException(modname, subname, exc)

    try:
        return getattr(getattr(pypkg, subname), classname)
    except Exception as exc:
        raise InvalidClassException(modname, subname, classname, exc)


def _splittype(typestring):
    """ Split the string the / delimiter and strip out empty strings

    Performs similar logic to roslib.names.package_resource_name but is a bit
    more forgiving about excess slashes
    """
    splits = [x for x in typestring.split("/") if x]
    if len(splits) == 2:
        return splits
    raise InvalidTypeStringException(typestring)


def _add_to_cache(cache, lock, key, value):
    lock.acquire()
    cache[key] = value
    lock.release()


def _get_from_cache(cache, lock, key):
    """ Returns the value for the specified key from the cache.
    Locks the lock before doing anything. Returns None if key not in cache """
    lock.acquire()
    ret = None
    if key in cache:
        ret = cache[key]
    lock.release()
    return ret


#
# License: BSD
#   https://raw.github.com/robotics-in-concert/rocon_tools/license/LICENSE
#

##############################################################################
# Description
##############################################################################

"""
.. module:: exceptions
   :platform: Unix
   :synopsis: Exceptions raised by loading/unloading of interactions.


This module defines exceptions raised by the rocon_interactions package.
These exception names are all included in the main
rocon_interactions namespace.  To catch one, import it this
way:

.. code-block:: python

    from rocon_interactions import InvalidInteraction

----

"""

##############################################################################
# Exceptions
##############################################################################


class InvalidInteraction(Exception):
    """
      Whenever an interaction has been specified incorrectly.
    """


class YamlResourceNotFoundException(IOError):
    """
      The requested yaml resource could not be found.
    """


class MalformedInteractionsYaml(Exception):
    """
      Whenever malformed yaml is used in loading a set of interactions.
    """


class FailedToStartRappError(Exception):
    """ Failed to start rapp. """
    pass


class FailedToStopRappError(Exception):
    """ Failed to stop rapp. """
    pass

class FailedToListRappsError(Exception):
    """ Failed to list rapps. """
    pass

#TODO : generic exceptions for any kind of message format


# -*- coding: utf-8 -*-
# This python package is handling all MOCK communication for rostful-node.
# The point of it is to be able to fully tests the multiprocess behavior,
#     in pure python, without having to run a ROS system.
#
# If possible these should also be base classes for all the possible multiprocess framework interfaces
# available in pyros
#
# Mock interface is provided with the package, because it may help client libraries to implement their own tests
#
from __future__ import absolute_import

from .transient_if import TransientIf
from .transient_if_pool import TransientIfPool, DiffTuple

__all__ = [
    'DiffTuple',
    'TransientIf',
    'TransientIfPool',
]



#!/usr/bin/env python
from __future__ import absolute_import

import six

import os
import sys

# This is needed if running this test directly (without using nose loader)
# prepending because ROS relies on package dirs list in PYTHONPATH and not isolated virtualenvs
# And we need our current module to be found first.
current_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..', '..'))
# if not current_path in sys.path:
sys.path.insert(1, current_path)  # sys.path[0] is always current path as per python spec

# Unit test import ( will emulate ROS setup if needed )
import time
import Queue

from pyros.rosinterface.subscriber_if import SubscriberBack
from pyros.rosinterface.message_conversion import get_msg, get_msg_dict, populate_instance, extract_values, FieldTypeMismatchException


# ROS imports should now work from ROS or from python (without ROS env setup)
import rospy
import roslaunch
import rostopic
import std_msgs.msg as std_msgs

import pyros

from pyros_utils import rostest_nose
import logging
import inspect
import unittest
import nose
from nose.tools import assert_false, assert_true, assert_equal


# test node process not setup by default (rostest dont need it here)
pub_process = None
echo_process = None


# This should have the same effect as the <name>.test file for rostest.
# Should be used only by nose ( or other python test tool )
def setup_module():
    if not rostest_nose.is_rostest_enabled():
        rostest_nose.rostest_nose_setup_module()

        # we still need a node to interact with topics
        rospy.init_node('TestStringTopic', anonymous=True, disable_signals=True)
        # CAREFUL : this should be done only once per PROCESS
        # Here we enforce TEST RUN 1<->1 MODULE 1<->1 PROCESS. ROStest style.


def teardown_module():
    if not rostest_nose.is_rostest_enabled():
        rostest_nose.rostest_nose_teardown_module()


# TODO : review this test (and ROS topic implementation) :
# The required functionality is to get the latest message on the topic
# But ideally we should never lose message on topic,
#   or behavior might be very different if message changes while one publisher spams it ( like here )
#  -> means we need an explicit way of consuming message
#  -> means we need paging to be able to access multiple messages
#  -> means we need memory and max reached behavior...
#  -> means we probably want to filter what message to keep or not
#  -> means we probably need an option to RLE compress messages in very fast topic rates ( "blahblah" arrived 5 times )
#  -> means we probably need a way to keep but greatly compress old messages ( logrotate style )


class TestStringSubscriber(unittest.TestCase):
    """ Testing the TopicBack class with String message """
    # misc method
    def logPoint(self):
        currentTest = self.id().split('.')[-1]
        callingFunction = inspect.stack()[1][3]
        print('in {0!s} - {1!s}()'.format(currentTest, callingFunction))

    def sub_cb(self, data):
        res = extract_values(data)
        self.msg_queue.put(res)

    def msg_extract(self, retries=5, retrysleep=1):
        msg = None
        try:
            msg = self.msg_queue.get_nowait()
        except Queue.Empty:
            pass
        retry = 0
        while not msg and retry < retries:
            print('Message not extracted. Retrying...')
            rospy.rostime.wallsleep(retrysleep)
            retry += 1
            try:
                msg = self.msg_queue.get_nowait()
            except Queue.Empty:
                pass
        if retry >= retries:
            self.fail('Message not extracted. Failing.')
        else:
            return msg

    def msg_wait(self, expected_msg, topic, retries=5, retrysleep=1):
        msg = self.msg_extract(topic)
        retry = 0
        # set XOR is a quick way to diff dict : if empty we stop (dicts are equals)
        while retry < retries and (set(six.iteritems(msg)) ^ set(six.iteritems(expected_msg))):
            print('Discarding unexpected message {0}. Retrying...'.format(msg))
            rospy.rostime.wallsleep(retrysleep)
            retry += 1
            msg = self.msg_extract(topic)
            print('Checking msg \"{0}\" == s \"{1}\"'.format(msg, expected_msg))

        if retry >= retries:
            self.fail('No Expected message {0} arrived. Failing.'.format(expected_msg))
        else:
            return msg

    def topic_wait_type(self, topic_name, retries=5, retrysleep=1):
        resolved_topic_name = rospy.resolve_name(topic_name)
        topic_type, _, _ = rostopic.get_topic_type(resolved_topic_name)
        topic_class, _, _ = rostopic.get_topic_class(resolved_topic_name)
        retry = 0
        while not topic_type and not topic_class and retry < 5:
            print('Topic {topic} not found. Retrying...'.format(topic=resolved_topic_name))
            rospy.rostime.wallsleep(retrysleep)
            retry += 1
            topic_type, _, _ = rostopic.get_topic_type(resolved_topic_name)
            topic_class, _, _ = rostopic.get_topic_class(resolved_topic_name)
        if retry >= retries:
            self.fail("Topic {0} not found ! Failing.".format(resolved_topic_name))
        return topic_type, topic_class

    def setUp(self):
        self.logPoint()
        """
        Non roslaunch fixture setup method
        :return:
        """
        # setting up internal queue to receive messages
        self.msg_queue = Queue.Queue()

        self.sub_topic_name = '/testing/Subscriber'

        # No need of parameter for that, any str should work
        self.test_message = "testing"

    def tearDown(self):
        self.logPoint()
        pass

    def test_subscriber_comm(self):
        try:
            self.logPoint()

            # exposing the topic for testing here
            self.sub_topic = rospy.Subscriber(self.sub_topic_name, std_msgs.String, self.sub_cb)
            # looking for the topic ( similar code than ros_interface.py )
            sub_topic_type, sub_topic_class = self.topic_wait_type(self.sub_topic_name)

            assert(sub_topic_class == std_msgs.String)
            self.sub_if = SubscriberBack(self.sub_topic_name, sub_topic_type)

            # Making sure the topic interface is ready to be used just after creation
            # by checking the number of connections on the actual subscriber
            pubs_connected = self.sub_topic.impl.get_num_connections()  # no local subs
            assert_true(pubs_connected == 1)

            # Topics are up. Use them.
            print("sending : {msg} on topic {topic}".format(msg=self.test_message, topic=self.sub_if.name))
            assert_true(self.sub_if.publish({'data': self.test_message}))

            print("waiting for : {msg} on topic {topic}".format(msg={'data': self.test_message}, topic=self.sub_topic.name))
            msg = self.msg_wait({'data': self.test_message}, self.sub_topic)
            # We assert there is no difference
            assert_true(len(set(six.iteritems(msg)) ^ set(six.iteritems({'data': self.test_message}))) == 0)

        except KeyboardInterrupt:
            self.fail("Test Interrupted !")


if __name__ == '__main__':
    print("ARGV : %r", sys.argv)
    # Note : Tests should be able to run with nosetests, or rostest ( which will launch nosetest here )
    rostest_nose.rostest_or_nose_main('pyros', 'testStringSubscriber', TestStringSubscriber, sys.argv)


from __future__ import absolute_import

import logging
from itertools import ifilter

import rostopic

from .api import rospy_safe as rospy

# create logger
_logger = logging.getLogger(__name__)
# and let it propagate to parent logger, or other handler
# the user of pyros should configure handlers

from ..baseinterface import TransientIfPool
from ..baseinterface import DiffTuple

from .topicbase import TopicTuple
from .publisher_if import PublisherBack
from .subscriber_if import SubscriberBack

try:
    import rocon_python_comms
except ImportError:
    rocon_python_comms = None




class RosPublisherIfPool(TransientIfPool):

    """
    MockInterface.
    """
    def __init__(self, publishers=None):
        # This base constructor assumes the system to interface with is already available ( can do a get_svc_available() )
        # CAREFUL publisher interfaces are subscribers
        super(RosPublisherIfPool, self).__init__(publishers, transients_desc="subscribers")

    def get_transients_available(self):  # function returning all services available on the system
        return self.available

    def transient_type_resolver(self, topic_name):  # function resolving the type of a service
        tpc = self.available.get(topic_name)
        if tpc:
            if tpc.type is None:  # if the type is unknown, lets discover it
                try:
                    resolved_topic_name = rospy.resolve_name(topic_name)
                    tpc.type, _, _ = rostopic.get_topic_type(resolved_topic_name)
                except rostopic.ROSTopicIOException:  # exception can occur -> just reraise
                   raise
            return tpc.type  # return the first we find. enough.
        else:
            rospy.logerr("ERROR while resolving {topic_name}. Topic not known as available. Ignoring".format(**locals()))
            return None

    def TransientMaker(self, topic_name, topic_type, *args, **kwargs):  # the service class implementation
        return PublisherBack(topic_name, topic_type, *args, **kwargs)

    def TransientCleaner(self, topic):  # the topic class cleanup implementation
        return topic.cleanup()

    ## bwcompat
    # REQUESTED
    @property
    def publishers_args(self):
        return self.transients_args

    # AVAILABLE
    @property
    def publishers_available(self):
        return self.available

    # INTERFACED
    @property
    def publishers(self):
        return self.transients

    # EXPOSE
    def expose_publishers(self, pub_regex):
        return self.expose_transients_regex(pub_regex)


    def get_publisher_list(self):  # function returning all services available on the system
        return self.get_transients_available

    def publisher_type_resolver(self, service_name):  # function resolving the type of a service
        return self.transient_type_resolver(self, service_name)

    def PublisherMaker(self, topic_name, topic_type):  # the publisher class implementation
        return self.TransientMaker(topic_name, topic_type)

    def PublisherCleaner(self, publisher):  # the publisher class cleanup implementation
        return self.TransientCleaner(publisher)
    ###########

    def reset_state(self, publishers, topic_types):
        """
        Reset internal system state representation.
        expect lists in format similar to masterAPI.
        :param topics
        :param topic_types:
        :return: a Difftuple, suitable to pass directly to
        """

        self.available = dict()
        for t in publishers:  # we assume t[1] is never empty here
            tt = next(ifilter(lambda ltt: t[0] == ltt[0], topic_types), [])
            ttp = TopicTuple(name=t[0], type=tt[1] if len(tt) > 0 else None, endpoints=set(t[1]))
            self.available[ttp.name] = ttp

        return publishers

    def compute_state(self, publishers_dt, topic_types_dt):
        """
        This is called only if there is a cache proxy with a callback, and expects DiffTuple filled up with names or types
        :param topics_dt:
        :return:
        """
        added_pubs = {t[0]: t[1] for t in publishers_dt.added}
        removed_pubs = {t[0]: t[1] for t in publishers_dt.removed}

        for apn, ap in added_pubs.iteritems():
            for rpn, rp in removed_pubs.iteritems():
                if rp in ap:  # remove nodes that are added and removed -> no change seen
                    ap.remove(rp)
                    removed_pubs[rpn].remove(rp)

        for rpn, rp in removed_pubs.iteritems():
            for apn, ap in added_pubs.iteritems():
                if ap in rp:  # remove nodes that are removed and added -> no change seen
                    rp.remove(ap)
                    added_pubs[apn].remove(ap)

        publishers_dt = DiffTuple(
            added=[[k, v] for k, v in added_pubs.iteritems()],
            removed=[[k, v] for k, v in removed_pubs.iteritems()]
        )
        computed_publishers_dt = DiffTuple([], [])
        _logger.debug("topics_dt : {publishers_dt}".format(**locals()))
        for t in publishers_dt.added:
            tt = next(ifilter(lambda ltt: t[0] == ltt[0], topic_types_dt.added), [])
            ttp = TopicTuple(name=t[0], type=tt[1] if len(tt) > 0 else None, endpoints=set(t[1]))
            if ttp.name in self.available:
                # if already available, we only update the endpoints list
                self.available[ttp.name].endpoints |= ttp.endpoints
            else:
                self.available[ttp.name] = ttp
                computed_publishers_dt.added.append(t[0])

        for t in publishers_dt.removed:
            tt = next(ifilter(lambda ltt: t[0] == ltt[0], topic_types_dt.removed), [])
            ttp = TopicTuple(name=t[0], type=tt[1] if len(tt) > 0 else None, endpoints=set(t[1]))
            if ttp.name in self.available:
                self.available[ttp.name].endpoints -= ttp.endpoints
                if not self.available[ttp.name].endpoints:
                    self.available.pop(ttp.name, None)
                    computed_publishers_dt.removed.append(t[0])

        # We still need to return DiffTuples
        return computed_publishers_dt

    def get_pub_interfaces_only_nodes(self):

        pubs_if = SubscriberBack.pool.get_all_interfaces()

        # inverting mapping for nodes ON and OFF separately:
        pubs_if_nodes_on = {}
        pubs_if_nodes_off = {}
        for node, data in pubs_if.iteritems():
            for t, ifon in data.get('publishers', {}).iteritems():  # we need to get the OTHER interfaces (pubs for subs, to prevent detection here)
                if ifon:
                    keys = pubs_if_nodes_on.setdefault(t, set())
                    # we also need to count the number of instance
                    # to avoid dropping Pub/Sub from tests or other annex code.
                    # SPECIAL CASE : normal pyros run should have only one reference to a Pub/sub
                    if PublisherBack.pool.get_impl_ref_count(t) <= 1:
                        keys.add(node)
                else:
                    keys = pubs_if_nodes_off.setdefault(t, set())
                    keys.add(node)

        return pubs_if_nodes_on, pubs_if_nodes_off

    # for use with line_profiler or memory_profiler
    # Not working yet... need to solve multiprocess profiling issues...
    # @profile
    def update_delta(self, publishers_dt, topic_types_dt=None):

        # FILTERING OUT TOPICS CREATED BY INTERFACE :

        # First we get all pubs/subs interfaces only nodes
        pubs_if_nodes_on, pubs_if_nodes_off = self.get_pub_interfaces_only_nodes()

        # print(" PUB ADDED DETECTED :")
        # print(publishers_dt.added)

        # Second we filter out ON interface topics from received ADDED topics list
        publishers_dt_added = [
            [t[0], [n for n in t[1]
                    if n not in pubs_if_nodes_on.get(t[0], set())
                    ]
             ]
            for t in publishers_dt.added
            ]

        # filtering out topics with no endpoints
        publishers_dt_added = [[tl[0], tl[1]] for tl in publishers_dt_added if tl[1]]

        # print(" PUB ADDED FILTERED :")
        # print(publishers_dt_added)
        #
        # print(" PUB REMOVED DETECTED :")
        # print(publishers_dt.removed)

        # Second we filter out OFF interface topics from received REMOVED topics list
        publishers_dt_removed = [
            [t[0], [n for n in t[1]
                    #if n not in topics_if_nodes_off.get(t[0], set())
                    # NOT DOABLE CURRENTLY : we would also prevent re interfacing a node that came back up...
                    # Probably better to fix flow between direct update and callback first...
                    # BUT compute_state() should take care of this...
                    ]
             ]
            for t in publishers_dt.removed
            ]

        # # Second we merge in ON interface topics into received REMOVED topics list
        # # This is useful to drop topics interfaces that are satisfying themselves...
        # for t, nodeset in pubs_if_nodes_on.iteritems():
        #     # note manipulating dictionaries will allow us to get rid of this mess
        #     found = False
        #     for td in publishers_dt_removed:
        #         if td[0] == t:
        #             td[1] += nodeset
        #             found = True
        #             break
        #     if not found:
        #         publishers_dt_removed.append([t, list(nodeset)])

        # filtering out topics with no endpoints
        publishers_dt_removed = [[tl[0], tl[1]] for tl in publishers_dt_removed if tl[1]]

        # print(" PUB REMOVED FILTERED :")
        # print(publishers_dt_removed)

        # computing state representation
        publishers_namelist_dt = self.compute_state(DiffTuple(
            added=publishers_dt_added,
            removed=publishers_dt_removed
        ), topic_types_dt or [])

        if publishers_namelist_dt.added or publishers_namelist_dt.removed:
            _logger.debug(
                rospy.get_name() + " Publishers Delta {publishers_namelist_dt}".format(**locals()))

        # TODO : put that in debug log and show based on python logger configuration

        # print("PUBLISHER APPEARED: {publishers_namelist_dt.added}".format(**locals()))
        # print("PUBLISHER GONE : {publishers_namelist_dt.removed}".format(**locals()))

        # update_services wants only names
        dt = self.transient_change_diff(
            transient_appeared=publishers_namelist_dt.added,
            transient_gone=publishers_namelist_dt.removed  # we want only hte name here
            # add_names=regexes_match_sublist(self.transients_args, [s[0] for s in topics_dt.added]),
            # remove_names=[s[0] for s in topics_dt.removed if s[0] not in self.get_transients_available()]
        )

        if dt.added or dt.removed:
            _logger.debug(
                rospy.get_name() + " Update Delta {dt}".format(**locals()))
            # print(" UPDATE DELTA:")
            # print(dt)
        return dt

    # for use with line_profiler or memory_profiler
    # Not working yet... need to solve multiprocess profiling issues...
    # @profile
    def update(self, publishers, topic_types):

        # FILTERING TOPICS CREATED BY INTERFACE :

        # First we get all pubs/subs interfaces only nodes
        pubs_if_nodes_on, pubs_if_nodes_off = self.get_pub_interfaces_only_nodes()

        #print("\n")
        #print(publishers)

        # Second we filter out ALL current and previous interface topics from received topics list
        publishers = [
            [t[0], [n for n in t[1]
                     if (n not in pubs_if_nodes_on.get(t[0], set())  # filter out ON interfaces to avoid detecting interface only topic
                         #n not in topics_if_nodes_off.get(t[0], set())  # filter out OFF interface to avoid re-adding interface that has been recently dropped
                         # NOT DOABLE CURRENTLY : we would also prevent re interfacing a node that came back up...
                         # Probably better to fix flow between direct update and callback first...
                         # BUT reset_state() should take care of this...
                        )
                    ]
            ]
            for t in publishers
        ]

        # TODO : maybe we need to reset the param once the dropped interface have been detected as droped (to be able to start the cycle again)

        #print("\n")
        #print(publishers)

        # filtering out topics with no endpoints
        publishers = [[tl[0], tl[1]] for tl in publishers if tl[1]]

        # First we need to reflect the external system state in internal cache
        self.reset_state(publishers, topic_types)

        #print("\nFULL TOPICS LIST: {topics}".format(**locals()))

        # Second we update our interfaces based on that new system state
        # TODO : pass full topic state here to avoid having to retrieve indirectly
        dt = self.transient_change_detect()

        return dt


TransientIfPool.register(RosPublisherIfPool)




from __future__ import absolute_import

from importlib import import_module
import re


def get_json_bool(b):
    if b:
        return 'true'
    else:
        return 'false'

def type_str(msg):
    return msg.__module__.split('.')[0] + '/' + msg.__name__

def load_type(msg_type_name):
    if msg_type_name.endswith('[]'):
        msg_type_name = msg_type_name[:-2]
    module_name, type_name = msg_type_name.split('/')
    msg_module = import_module(module_name + '.msg')
    # we should try to load the full service type if not found in msg (without needing Request or Response suffix)
    if not hasattr(msg_module,type_name):  # and (type_name.endswith('Request') or type_name.endswith('Response')):
        msg_module = import_module(module_name + '.srv')
    if not hasattr(msg_module,type_name):
        raise TypeError('Unknown ROS msg {0!s}'.format(msg_type_name))
    return getattr(msg_module,type_name)



from __future__ import absolute_import, print_function

import sys
import os
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..', '..')))

import nose
from nose.tools import assert_true, assert_false, assert_raises
import unittest

from pyros.rosinterface.mock import extract_values, populate_instance, FieldTypeMismatchException, NonexistentFieldException, StatusMsg
from pyros.rosinterface.mock import statusecho_service, MockSystem


class TestMockService(unittest.TestCase):
    """
    Main test fixture holding all tests
    Subclasses can override setup / teardown to test different environments
    """

    # Class fixture ( once each )
    @classmethod
    def setup_class(cls):
        cls.system = MockSystem()

    @classmethod
    def teardown_class(cls):
        pass

    def test_echo_status(self):
        msg = populate_instance({
            'error': 'BAD_ERROR',
            'code': 7,
            'message': 'A bad Error happened'
        }, StatusMsg('ERROR', 42, 'details'))
        assert_true(isinstance(msg, StatusMsg))
        echo_service = self.system.create_service('random_service', statusecho_service)
        resp = echo_service.call(msg)
        print(resp)
        assert_true(resp.error == 'BAD_ERROR')
        assert_true(resp.code == 7)
        assert_true(resp.message == 'A bad Error happened')

# TODO : make sure we handle errors when the service is exposed, but not existing any longer

if __name__ == '__main__':

    nose.runmodule()




from __future__ import absolute_import
from collections import namedtuple, deque

from ...baseinterface import TransientIf


class MockPublisher(TransientIf):
    """
    MockPublisher is a mock of the class handling conversion from Python API to Publisher call
    """

    def __init__(self, topic_name, topic_type):

        # getting the fullname to make sure we start with /
        topic_name = topic_name if topic_name.startswith('/') else '/' + topic_name

        topic_type = topic_type.msgtype

        super(MockPublisher, self).__init__(topic_name, topic_type)

    def cleanup(self):
        pass

    def publish(self, msg):
        # System should be set by system itself ( TODO : investigate proper way to inject dependency for Mock...)
        return self.system.transfer(msg, self.name)





# -*- coding: utf-8 -*-
#This python package is handling all ROS related communication for rostful-node.
from __future__ import absolute_import

import logging

"""
Hopefully this should endup in rosinterface.__doc__
"""

# create logger
# TODO solve multiprocess logger problem(s)...
_logger = logging.getLogger(__name__)
# and let it propagate to parent logger, or other handler
# the user of pyros should configure handlers

# Here we use pyros_setup configuration to setup ROS if needed before using any of our modules
# This is much simpler since we don't need any other configuration at import time.
# Extra Pyros configuration can be passed at runtime directly to the interface.

try:
    from .api import rospy_safe  # early except to prevent unintentional workaround in all modules here
    # We ideally should add all dependencies imported by the modules in this subpackage...
    # We should be fine here when running from deb.
    # But when running from python we have to except here to get environment setup properly in child process.
except ImportError:
    # This should be found from python dependencies. But not needed when running from deb.
    # Here should be the first time we setup ros variable in pyros package,
    # in the child process, since the child (and only the child) needs to know about ROS.
    _logger.warning("loading pyros_setup and configuring your ROS environment")
    import pyros_setup
    # This will load the pyros_setup configuration from the environment
    pyros_setup.configurable_import().configure().activate()
    # validate we can load ROS modules. Note other variables (like ROS_PACKAGE_PATH) should also be available.
    from .api import rospy_safe

from .service_if import ServiceBack
from .param_if import ParamBack
from .param_if_pool import RosParamIfPool
from .service_if_pool import RosServiceIfPool
from .subscriber_if_pool import RosSubscriberIfPool
from .publisher_if_pool import RosPublisherIfPool
from .baseinterface import BaseInterface
from .ros_interface import RosInterface


__all__ = [
    'TopicBack',
    'ServiceBack',
    'ParamBack',
    'RosInterface',
]


from __future__ import absolute_import

import os
import sys
import pickle

# This is needed if running this test directly (without using nose loader)
# prepending because ROS relies on package dirs list in PYTHONPATH and not isolated virtualenvs
# And we need our current module to be found first.
current_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..', '..'))
# if not current_path in sys.path:
sys.path.insert(1, current_path)  # sys.path[0] is always current path as per python spec

# Unit test import
from pyros.rosinterface import message_conversion as msgconv

# ROS imports should now work from ROS or from python ( even without ROS env setup)
import rospy

# useful test tools
import nose
from nose.tools import assert_equal, assert_true, assert_false

# Test all standard message
import std_msgs.msg as std_msgs


# TODO : change this customJSON -> ROS conversion by JSON -ujson-> py -rospy-> ROS
# TODO : achieve SYMMETRY : what we put in == what we get out ( no "data" field added )
def test_String_default():
    msg = std_msgs.String()
    val = msgconv.extract_values(msg)
    assert_equal(val["data"], str())  # "data" : should not appear here


def test_String_custom():
    msg = std_msgs.String("teststr")
    msgconv.populate_instance({"data": "teststr2"}, msg)  # we shouldnt need "data" here
    val = msgconv.extract_values(msg)
    assert_equal(val["data"], "teststr2")


def test_msg_exception_pickle():
    exc = msgconv.NonexistentFieldException("message type", ["field1", "field2"])

    pbuf = pickle.dumps(exc)
    pexc = pickle.loads(pbuf)

    assert_equal(pexc.basetype, "message type")
    assert_equal(len(pexc.fields), 2)
    assert_true("field1" in pexc.fields)
    assert_true("field2" in pexc.fields)

# TODO : assert both "string" and "std_msgs/String" are convertible from&to "str" and "unicode"


#TODO : assert exception are being thrown

if __name__ == '__main__':
    nose.runmodule()



#!/usr/bin/python
# All ways to run pyros and all Manager commands
# should be defined here for consistency
from __future__ import absolute_import
import os
import sys

import six

import click

# logging configuration should be here to not be imported by python users of pyros.
# only used from command line

import logging
import logging.config
# Setting up logging for this test
logging.config.dictConfig(
    {
        'version': 1,
        'formatters': {
            'verbose': {
                'format': '%(levelname)s %(asctime)s %(module)s %(process)d %(thread)d %(message)s'
            },
            'simple': {
                'format': '%(levelname)s %(name)s:%(message)s'
            },
        },
        'handlers': {
            'null': {
                'level': 'DEBUG',
                'class': 'logging.NullHandler',
            },
            'console': {
                'level': 'DEBUG',
                'class': 'logging.StreamHandler',
                'formatter': 'simple'
            },
        },
        'loggers': {
            'pyros_config': {
                'handlers': ['console'],
                'level': 'INFO',
                'propagate': False,
            },
            'pyros_setup': {
                'handlers': ['console'],
                'level': 'INFO',
            },
            'pyros': {
                'handlers': ['console'],
                'level': 'INFO',
            }
        }
    }
)


import nose
import pkg_resources

_path = pkg_resources.resource_filename("pyros", "__main__.py")
_parent = os.path.normpath(os.path.join(os.path.dirname(_path), ".."))


# importing current package if needed ( solving relative package import from __main__ problem )
# if __package__ is None:
#     sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
#     from rostful import create_app, set_pyros_client, setup_app_routes
# else:
#     from . import create_app, set_pyros_client, setup_app_routes


def nosemain():
    args = sys.argv + [opt for opt in (
        # "--exe",  # DO NOT look in exe (maybe old rostests ?)
        # "--all-modules",  # DO NOT look in all modules
        "--traverse-namespace",
        "--verbosity=2",
        "--with-id",
        "--with-xunit",
        "--where=%s" % _parent
    )
    if opt not in sys.argv
    ]
    nose.run(argv=args)


def pyros_rosinterface_launch(node_name=None, pyros_config=None, ros_argv=None):

    node_name = node_name or 'pyros_rosinterface'
    ros_argv = ros_argv or []

    # dynamic setup and import
    try:
        import pyros
        node_proc = pyros.PyrosROS(
            node_name,
            ros_argv
        )
        # Attribute error is triggered when using pyros < 0.1.0
    except (ImportError, AttributeError) as e:
        #logging.warn("{name} Error: Could not import pyros : {e}".format(name=__name__, e=e))
        ### TMP ###
        logging.warn("{name} Attempting bwcompat import : {e}".format(name=__name__, e=e))
        # BWcompat with old pyros :
        try:
            # this will import rosinterface and if needed simulate ROS setup
            import sys
            import pyros
            import pyros.rosinterface
            # this doesnt work with celery handling of imports
            # sys.modules["pyros.rosinterface"] = pyros.rosinterface.delayed_import_auto(
            #     distro='indigo',
            #     base_path=os.path.join(os.path.dirname(__file__), '..', '..', '..')
            # )
            pyros.rosinterface = pyros.rosinterface.delayed_import_auto(
                distro='indigo',
                base_path=os.path.join(os.path.dirname(__file__), '..', '..', '..')
            )

            node_proc = pyros.rosinterface.PyrosROS(
                node_name,
                ros_argv,
                base_path=os.path.join(os.path.dirname(__file__), '..', '..', '..')
            )

            node_proc.configure(pyros_config)

        except ImportError as e:
            logging.warn("{name} Error: Could not import pyros.rosinterface : {e}".format(name=__name__, e=e))
            raise

    return node_proc


# Change that into something cleaner and related with the app itself (not server)
# Middleware ? app context ? Tricky with dynamic imports...
# middleware ref : https://github.com/miguelgrinberg/python-engineio/blob/master/engineio/middleware.py#L49
def pyros_start(config, ros_args='', pyros_ctx_impl=None):

    # implementing Config.get_namespace() for flask version < 1.0:
    namespace = 'PYROS_'
    trim_namespace = True
    lowercase = False
    rv = {}
    for k, v in six.iteritems(config):
        if not k.startswith(namespace):
            continue
        if trim_namespace:
            key = k[len(namespace):]
        else:
            key = k
        if lowercase:
            key = key.lower()
        rv[key] = v
    # rv should now contain a dictionary of namespaced key:value from config

    try:
        from pyros import pyros_ctx, PyrosClient
    except Exception as e:
        logging.error("pyros module is not accessible in sys.path. It is required to run rostful.", exc_info=True)
        logging.error("sys.path = {0}".format(sys.path))
        raise

    # default to real module, if no other implementation passed as parameter (used for mock)
    pyros_ctx_impl = pyros_ctx_impl or pyros_ctx

    # One PyrosNode is needed for Flask.
    # TODO : check everything works as expected, even if the WSGI app is used by multiple processes

    try:
        node_ctx_gen = pyros_ctx_impl(name='rostful', argv=ros_args, pyros_config=rv)
    except TypeError as te:
        # bwcompat trick
        node_ctx_gen = pyros_ctx_impl(name='rostful', argv=ros_args,
                            base_path=os.path.join(os.path.dirname(__file__), '..', '..', '..'))
    return node_ctx_gen


# TODO : handle ros arguments here
# http://click.pocoo.org/5/commands/#group-invocation-without-command
@click.group()
def cli():
    pass


@cli.command()
def test():
    errno = nosemain()
    sys.exit(errno)


#
# Arguments' default value is None here
# to use default values from config file if one is provided.
# If no config file is provided, internal defaults are used.
#
@cli.command()
@click.option('-i', '--interface', default='ros', type=click.Choice(['ros', 'ros_mock']))
# @click.option('-a', '--async', default=False)  # wether to activate async implementation (instead of pyros main loop)
@click.option('-c', '--config', default=None)  # this is the last possible config override, and has to be explicit.
@click.option('-l', '--logfile', default=None)  # this is the last possible logfile override, and has to be explicit.
@click.option('ros_args', '-r', '--ros-arg', multiple=True, default='')
def run(interface, config, logfile, ros_args):
    """
    Start a pyros node.
    :param interface: the interface implementation (ROS, Mock, ZMP, etc.)
    :param config: the config file path, absolute, or relative to working directory
    :param logfile: the logfile path, absolute, or relative to working directory
    :param ros_args: the ros arguments (useful to absorb additional args when launched with roslaunch)
    """
    logging.info(
        'pyros started with : interface {interface} config {config} logfile {logfile} ros_args {ros_args}'.format(
            interface=interface, config=config, logfile=logfile, ros_args=ros_args))

    if interface == 'ros':
        node_proc = pyros_rosinterface_launch(node_name='pyros_rosinterface', pyros_config=config, ros_argv=ros_args)
    else:
        node_proc = None  # NOT IMPLEMENTED

    # node_proc.daemon = True  # we do NOT want a daemon(would stop when this main process exits...)
    client_conn = node_proc.start()  # in a sub process
    # DISABLING THIS FOR NOW, process tree is a bit unexpected...
    # TODO : investigate
    # client_conn = node_proc.run()  # in same process


if __name__ == '__main__':
   cli()



import re
from collections import namedtuple, OrderedDict, Iterable
from io import StringIO
import inspect
import json

def _sec_name_to_dfn(name):
    idx = name.find('Section')
    if idx == -1:
        return name + 'Definition'
    return re.sub(r'Section','Definition',name)

def _clean_string(s):
    if s is None:
        s = ''
    elif isinstance(s,(list, tuple, dict)):
        s = json.dumps(s)
    else:
        s = str(s)
    return s

class ParsingError(Exception):
    pass

class MissingDefinitionError(Exception):
    pass

class DefFile(object):
    def __init__(self, format=None, manifest=True, definitions=None, sections=None):
        self.format = format
        if manifest is True:
            self.manifest = Manifest()
        elif manifest:
            self.manifest = manifest
        else:
            self.manifest = None
        self.definitions = [d for d in definitions or []]
        self.sections = ordereddicti(sections or {})
    
    @property
    def type(self):
        if not self.manifest:
            return None
        return self.manifest.def_type
    
    @type.setter
    def type(self, value):
        if not self.manifest:
            self.manifest = Manifest()
        self.manifest.def_type = value
    
    def definition_types(self):
        types = set()
        for definition in self.definitions:
            types.add(definition.type)
        return list(types)
    
    def has_definition(self, dfn_type, name):
        for d in self.definitions:
            if d.type == dfn_type and d.name == name:
                return True
        return False
    
    def get_definition(self, dfn_type, name):
        for d in self.definitions:
            if d.type == dfn_type and d.name == name:
                return d
        return None
    
    def get_definitions_of_type(self, dfn_type):
        return [d for d in self.definitions if d.type == dfn_type]
    
    def add_definition(self, dfn, quiet=False):
        for d in self.definitions:
            if d.type == dfn.type and d.name == dfn.name:
                if quiet:
                    return False
                else:
                    raise TypeError("{0!s}:{1!s} definition already exists".format(dfn.type, dfn.name))
        self.definitions.append(dfn)
        if quiet:
            return True
    
    def has_section(self, section, format=None):
        return _clean_string(section) in self.sections
    
    def get_section(self, section, format=None):
        sec = self.sections.get(_clean_string(section))
        if sec and format and sec.format != format:
            return None
        else:
            return sec
    
    def add_section(self, section):
        self.sections[_clean_string(section.name)] = section
    
    def create_sections(self, section_type, *args):
        if len(args) == 1 and isinstance(args[0],(list,tuple)):
            names = args[0]
        else:
            names = args
        for name in names:
            section = section_type(name)
            self.sections[section.name] = section
    
    def load_includes(self, path=None):
        pass
    
    def tojson(self):
        d = {}
        
        d['Manifest'] = self.manifest.tojson()
        
        d['Definitions'] = dict([(dfn.type + ':' + dfn.name, dfn.tojson()) for dfn in self.definitions])
        
        d['Sections'] = dict([(section_name, section.tojson()) for section_name, section in self.sections.iteritems()])
        
        return d
    
    def tostring(self, suppress_formats=False):
        strs = []
        if self.manifest:
            s = ''
            if self.format and not suppress_formats:
                s += '[Manifest!{0!s}]\n'.format(self.format)
            else:
                s += '[Manifest]\n'
            s += self.manifest.__str__()
            strs.append(s)
        elif self.format and not suppress_formats:
            strs.append('[Manifest!{0!s}]'.format(self.format))
        
        for dfn in self.definitions:
            s = ''
            if dfn.format and not suppress_formats:
                s += '[{0!s}:{1!s}!{2!s}]\n'.format(dfn.type, dfn.name, dfn.format)
            else:
                s += '[{0!s}:{1!s}]\n'.format(dfn.type, dfn.name)
            s += dfn.__str__()
            strs.append(s)
        
        for sec_name, section in self.sections.iteritems():
            s = ''
            sec_name = _clean_string(section.name)
            if sec_name.find(':') != -1:
                sec_name = ':' + sec_name
            if section.format and not suppress_formats:
                s += '[{0!s}!{1!s}]\n'.format(sec_name, section.format)
            else:
                s += '[{0!s}]\n'.format(sec_name)
            s += section.__str__()
            strs.append(s)
        
        return '\n\n'.join(strs)
    
    def get(self, key, default=None):
        if isinstance(key, tuple):
            if len(key) == 1:
                key = (None,key[0])
            sec = self.get_section(key[0])
            if not sec:
                return default
            if len(key) == 2:
                return sec.__getitem__(key[1])
            else:
                return sec.__getitem__(key[1:])
        else:
            null_sec_val = self.get_section(None).__getitem__(key) if self.has_section(None) and hasattr(self.get_section(None),'__getitem__') else None
            if (self.manifest and key in self.manifest 
                and null_sec_val is not None):
                raise RuntimeError('Manifest and null section both have key {0!s}!'.format(str(key)))
            elif null_sec_val is not None:
                return null_sec_val
            elif not self.manifest:
                return default
            else:
                return self.manifest.get(key,default)
    
    def get_bool(self, key, default=False):
        value = self.get(key, default=None)
        if value is None:
            return default
        elif isinstance(value, basestring):
            return value.lower() in ['true','1']
        else:
            return False
    
    def get_int(self, key, default=None):
        value = self.get(key, default=None)
        if value is None:
            return default
        try:
            return int(value)
        except:
            return None
    
    def get_list(self, key, split_pattern=', ', default=None, strip_brackets=True, json=True):
        if default is None:
            default = []
        val = self.get(key)
        if val is None:
            return default
        
        if json:
            import json as jsonmod
            try:
                json_value = jsonmod.loads(val)
                if isinstance(json_value, list):
                    return json_value
            except:
                pass
        
        if strip_brackets and ((val.startswith('[') and val.endswith(']'))
                                or (val.startswith('(') and val.endswith(')'))):
            val = val[1:-1]
        return [v for v in re.split(split_pattern, val) if v]
    
    def __contains__(self, key):
        if isinstance(key, tuple):
            if len(key) == 1:
                key = (None,key[0])
            sec = self.get_section(key[0])
            return key in sec
        else:
            null_sec_val = None
            if self.has_section(None):
                try:
                    null_sec_val = key in self.get_section(None)
                except:
                    pass
            if (self.manifest and key in self.manifest 
                and null_sec_val is True):
                raise RuntimeError('Manifest and null section both have key {0!s}!'.format(str(key)))
            elif null_sec_val is not None:
                return null_sec_val
            elif not self.manifest:
                return False
            else:
                return key in self.manifest
    
    def __getitem__(self, key):
        return self.get(key)
    
    def __str__(self):
        return self.tostring()
    
    def __repr__(self):
        return 'DefFile(format={0!s},manifest={1!s},definitions={2!s},sections={3!s})'.format(repr(self.format), repr(self.manifest), repr(self.definitions), repr(self.sections))
                

class Manifest(object):
    def __init__(self, fields=None, **kwargs):
        self.def_type = kwargs.get('def_type')
        self.includes = kwargs.get('includes',[])
        self.fields = strstrdicti(fields or {})
        
    def tojson(self):
        d = dict(self.fields.iteritems())
        if self.def_type is not None:
            d['Def-Type'] = self.def_type
        return d
    
    def tostring(self):
        strs = []
        if self.def_type is not None:
            strs.append('Def-Type = {0!s}'.format(self.def_type))
        for include_file, include_type in self.includes:
            if include_type:
                strs.append('Include.{0!s} = {1!s}'.format(include_type, include_file))
            else:
                strs.append('Include = {0!s}'.format(include_file))
        for key, value in self.fields.iteritems():
            strs.append('{0!s} = {1!s}'.format(key, value))
        return '\n'.join(strs)
    
    def __str__(self):
        return self.tostring()
    
    def __repr__(self):
        def_type_str = ',{0!s}'.format(repr(self.def_type)) if self.def_type is not None else ''
        includes_str = ',{0!s}'.format(repr(self.includes)) if self.includes else ''
        return 'Manifest(fields={0!s}{1!s}{2!s})'.format(repr(self.fields), def_type_str, includes_str)
    
    def __iter__(self):
        return self.fields.iteritems()
    
    def __nonzero__(self):
        return bool(self.fields or self.def_type or self.includes)
    
    def __contains__(self, key):
        return self.fields.__contains__(key)
  
    def __getitem__(self, key):
        """If the section has the key, return a (possibly empty) string.
        Otherwise, return None."""
        if self.fields.has_key(key):
            return self.fields.get(key)
        else:
            return None
  
    def __setitem__(self, key, value):
        self.fields[key] = value

    def get(self, key, default=None):
        return self.fields.get(key, default)

    def has_key(self, key):
        return self.fields.has_key(key)

    def items(self):
        return [(k, v) for k, v in self.iteritems()]
    
    def keys(self):
        return [k for k in self.iterkeys()]
    
    def values(self):
        return [v for v in self.itervalues()]
    
    def iteritems(self):
        return self.fields.iteritems()
        
    def iterkeys(self):
        return self.fields.iterkeys()
        
    def itervalues(self):
        return self.fields.itervalues()
    
    def __setattr__(self, name, value):
        if name == 'fields':
            value = strstrdicti(value) if value is not None else strstrdicti()
        return super(Manifest,self).__setattr__(name, value)

class Section(object):
    FORMAT = None
    def __init__(self,name):
        self.name = name
    
    @property
    def format(self):
        if not hasattr(self, '_format'):
            return self.FORMAT
        else:
            return self._format
    @format.setter
    def format(self,value):
        setattr(self,'_format',value)
    
    def tojson(self):
        raise NotImplementedError()
    
    def tostring(self):
        raise NotImplementedError()
    
    def __str__(self):
        return self.tostring()

class Definition(Section):
    FORMAT = None
    
    def __init__(self,type,name):
        super(Definition,self).__init__(name)
        self.type = type
    
    @property
    def format(self):
        if not hasattr(self, '_format'):
            return self.FORMAT
        else:
            return self._format
    @format.setter
    def format(self,value):
        setattr(self,'_format',value)
    
    def tojson(self):
        raise NotImplementedError()
    
    def tostring(self):
        raise NotImplementedError()
    
    def __str__(self):
        return self.tostring()

class INISection(Section):
    FORMAT = 'ini'
    def __init__(self, name, fields=None, discard_none=False):
        Section.__init__(self, name)
        self.fields = strstrdict((key, value) for key, value in fields.iteritems() if not (value is None and discard_none)) if fields else strstrdict()
    
    def tojson(self):
        return self.fields.copy()
    
    def tostring(self):
        strs = []
        for key, value in self.fields.iteritems():
            if value.find('\n') != -1:
                strs.append("{0!s} = |".format(key))
                [strs.append(re.sub(r'=',r'\\=',line)) for line in value.split('\n')]
            else:
                strs.append('{0!s} = {1!s}'.format(key, value))
        return '\n'.join(strs)
    
    @staticmethod
    def load_ini_dict(section_dict):
        ini = {}
        for section_name, section in section_dict.iteritems():
            if not isinstance(section, INISection): continue
            for key, value in section.fields:
                ini[section_name + '.' + key] = value
        return ini
    
    def __repr__(self):
        return 'INISection({0!s}, fields={1!s})'.format(repr(self.name), repr(self.fields))
    
    def __iter__(self):
        return self.fields.iteritems()
    
    def __nonzero__(self):
        return bool(self.fields)
    
    def __contains__(self, key):
        return key in self.fields
  
    def __getitem__(self, key):
        """If the section has the key, return a (possibly empty) string.
        Otherwise, return None."""
        if self.fields.has_key(key):
            return self.fields.get(key)
        else:
            return None
  
    def __setitem__(self, key, value):
        self.fields[key] = value

    def get(self, key, default=None):
        return self.fields.get(key, default)

    def has_key(self, key):
        return self.fields.has_key(key)

    def items(self):
        return [(k, v) for k, v in self.iteritems()]
    
    def keys(self):
        return [k for k in self.iterkeys()]
    
    def values(self):
        return [v for v in self.itervalues()]
    
    def iteritems(self):
        return self.fields.iteritems()
        
    def iterkeys(self):
        return self.fields.iterkeys()
        
    def itervalues(self):
        return self.fields.itervalues()
    
    def __setattr__(self, name, value):
        if name == 'fields':
            value = strstrdict(value) if value is not None else strstrdict()
        return super(INISection,self).__setattr__(name, value)

class RawSection(Section):
    FORMAT = 'raw'
    def __init__(self, name, content=None):
        Section.__init__(self,name)
        self.content = str(content) if content else ''
    
    def __nonzero__(self):
        return bool(self.content)
    
    def tojson(self):
        self.content
    
    def tostring(self):
        return self.content
    
    def __repr__(self):
        return 'RawSection({0!s},content={1!s})'.format(repr(self.name), repr(self.content))

def get_definition_from_section(section_class):
    sec_class_name = section_class.__name__
    dfn_class_name = _sec_name_to_dfn(sec_class_name)
    class DfnClass(section_class,Definition):
        def __init__(self,type,name, *args, **kwargs):
            Definition.__init__(self,type,name)
            section_class.__init__(self,name,*args,**kwargs)
        
        def __repr__(self):
            sec_repr = section_class.__repr__(self)
            return sec_repr.replace('{0!s}('.format(sec_class_name),'{0!s}({1!s},'.format(dfn_class_name, repr(self.type)))
    DfnClass.__name__ = dfn_class_name
    return DfnClass

INIDefinition = get_definition_from_section(INISection)

class ROSStyleDefinition(Definition):
    Field = namedtuple('Field', ['name','type'])
    FORMAT = 'ros'
    def __init__(self,type,name,segment_names,segment_values=None):
        super(ROSStyleDefinition,self).__init__(type,name)
        
        self.segment_names = segment_names
        if segment_values:
            self.segments = tuple(segment_values)
        else:
            self.segments = tuple([[] for _ in xrange(len(segment_names))])
    
    def segment(self,index):
        if not isinstance(index,basestring):
            return self.segments[index]
        for idx, seg_name in enumerate(self.segment_names):
            if seg_name == index:
                return self.segment(idx)
        raise IndexError("No segment named {0!s}".format(index))
             
    def __getitem__(self,index):
        if len(self.segments) == 1:
            return self.segments[0][index]
        else:
            return self.segment(index)
    
    def tojson(self):
        segs = []
        for segment in self.segments:
            segs.append(list(segment))
        if len(self.segments) == 1:
            return segs[0]
        else:
            return dict(zip(self.segment_names,segs))
    
    def tostring(self):
        segment_strs = []
        for segment in self.segments:
            strs = []
            for name, type in segment:
                strs.append('{0!s} {1!s}'.format(type, name))
            segment_strs.append('\n'.join(strs))
        return '\n----\n'.join(segment_strs)
    
    def __repr__(self):
        return 'ROSStyleDefinition({0!s},{1!s},{2!s},segment_values={3!s})'.format(repr(self.type), repr(self.name), repr(self.segment_names), repr(self.segments))


class DefFileParser(object):
    DEF_TYPE_PATT = r'(?P<dfn_type>[a-zA-Z_]\w*)'
    SEC_NAME_PATT = r'(?P<section>(?:[^][\\!#]|(?:\\[][\\!#]))+)'
    FORMAT_PATT = r'(?P<format>[a-zA-Z_]\w*)'
    COMMENT_PATT = r'(?:#(?P<comment>.*))?'
    
    SECTION_PATT = r'\[\s*(?:(?:{dfn_type})?\s*:\s*)?{section}(?:\s*!\s*{format})?\s*\]'.format(
            dfn_type=DEF_TYPE_PATT,section=SEC_NAME_PATT,format=FORMAT_PATT)
    EMPTY_SECTION_PATT = r'\[\s*(?:\s*!\s*{format})?\s*\]'.format(
            format=FORMAT_PATT)
    SECTION_RE = re.compile(SECTION_PATT)
    EMPTY_SECTION_RE = re.compile(EMPTY_SECTION_PATT)
    
    KEY_VALUE_PATT_TEMPLATE = r'^(?P<DefFileParser_key>{key_patt})\s*(?<!\\)=\s*(?P<DefFileParser_value>{value_patt})$'
    
    BASIC_KEY_CHAR_PATT = r'(?:[^ \t\n\r\f\v=]|\\=)'
    BASIC_KEY_PATT = r'{char}+(?:\s+{char}+)*'.format(char=BASIC_KEY_CHAR_PATT) #r'[^\s=](?:\s*\S)*'
    BASIC_VALUE_PATT = r'(?:\S(?:\s*\S)*)?'
    BASIC_KEY_VALUE_PATT = KEY_VALUE_PATT_TEMPLATE.format(key_patt=BASIC_KEY_PATT,value_patt=BASIC_VALUE_PATT)
    BASIC_KEY_VALUE_RE = re.compile(BASIC_KEY_VALUE_PATT)
    
    @classmethod
    def get_section_data(cls,line):
        match = cls.SECTION_RE.match(line) or cls.EMPTY_SECTION_RE.match(line)
        if not match:
            return None
        else:
            data = match.groupdict(None)
            data['section'] = re.sub(r'\\(?!\\)','',data.get('section',''))
            data['dfn_type'] = data.get('dfn_type')
        return data
    
    @classmethod
    def is_section_line(cls,line):
        return bool(cls.SECTION_RE.match(line))
    
    def __init__(self,default_format=None, default_def_type=None, require_def_type=None):
        self.default_format = default_format
        self.default_def_type = default_def_type
        self.require_def_type = require_def_type or default_def_type
        self.null_section_parser = None
        
        self.definition_parsers = {}
        self.section_parsers = {}
        
        self.fp = None
        self.current_line = 0
        self.null_section = True
    
    def add_section_parser(self,parser, name, format, def_file_format=None):
        self.section_parsers[(name,format,def_file_format)] = parser
    
    def add_default_section_parser(self,parser = None, def_file_format=None, null_section=False):
        parser = parser or RawSectionParser
        self.section_parsers[(None,None,def_file_format)] = parser
        if null_section:
            self.set_null_section_parser(parser)
    
    def set_null_section_parser(self,parser):
        self.null_section_parser = parser
    
    def get_section_parser(self,name,format,def_file_format,go_easy=False):
        matches = {}
        for key, parser in self.section_parsers.iteritems():
            parser_name, parser_format, parser_def_file_format = key
            if (parser_name and parser_name != name) or \
                    (parser_format and parser_format != format) or \
                    (parser_def_file_format and parser_def_file_format != def_file_format):
                continue
            score = 0
            if parser_format == format:
                score +=1
            if parser_name:
                score += 2
            matches[score] = parser
        if not matches:
            if go_easy:
                return None
            elif format:
                raise ParsingError('No section parser for section {0!s} and format {1!s}'.format(name, format))
            else:
                raise ParsingError('No section parser for section {0!s}'.format(name))
        return matches[max(matches.keys())](name,format,self._get_reader())
    
    def add_definition_parser(self,parser, dfn_type, format, def_file_format=None):
        if not dfn_type:
            raise TypeError('dfn_type cannot be None!')
        self.definition_parsers[(dfn_type,format,def_file_format)] = parser
    
    def get_definition_parser(self,dfn_type,name,format,def_file_format,go_easy=False):
        matches = {}
        for key, parser in self.definition_parsers.iteritems():
            parser_def_type, parser_format, parser_def_file_format = key
            if (parser_def_type != dfn_type) or \
                    (parser_format and parser_format != format) or \
                    (parser_def_file_format and parser_def_file_format != def_file_format):
                continue
            score = 0
            if parser_format == format:
                score +=1
            matches[score] = parser
        if not matches:
            parser = self.get_section_parser(None, format, def_file_format, go_easy=True)
            if parser:
                return DefFileParser.DefinitionParser.from_section_parser(parser)
        
        if not matches:
            if go_easy:
                return None
            elif format:
                raise ParsingError('No definition parser for type {0!s} and format {1!s}'.format(dfn_type, format))
            else:
                raise ParsingError('No definition parser for type {0!s}'.format(dfn_type))
        return matches[max(matches.keys())](dfn_type,name,format,self._get_reader())
    
    def readline(self,new=True):
        if self.current_line == 0 or new:
            line = self.fp.readline()
            if not line:
                self.current_line = None
            else:
                if line.endswith('\n'):
                    line = line[:-1]
                self.current_line = re.sub(self.COMMENT_PATT,'',line).strip()
        if self.current_line == '':
            return self.readline()
        return self.current_line
    
    def sections(self):
        line = self.readline(new=False)
        while line is not None:
            section_info = self.get_section_data(line)
            if not section_info:
                if self.null_section and self.null_section_parser:
                    parser = self.null_section_parser(None,None,self._get_reader())
                    sec = parser.parse()
                    section_info = {'section': None, 'data': sec}
                else:
                    raise ParsingError('Parsing error: invalid section line\n{0!s}'.format(line))
            self.null_section = False
            yield section_info
            line = self.readline(new=False)
    
    def _get_reader(self):
        def reader():
            line = self.readline()
            while line and not DefFileParser.is_section_line(line):
                yield line
                line = self.readline()
        return reader
    
    def parse(self,fp,**kwargs):
        if isinstance(fp,basestring):
            if fp.find('\n') == -1:
                fp = open(fp)
            else:
                fp = StringIO(unicode(fp))
        self.fp = fp
        self.current_line = 0
        self.null_section = True
        default_format = kwargs.get('default_format',self.default_format)
        
        def_file = DefFile(manifest=False)
        for section in self.sections():
            if section['section'] is None:
                def_file.sections[''] = section['data']
            elif section['section'].lower() == 'manifest':
                if def_file.definitions or def_file.sections:
                    raise ParsingError('Parsing error: manifest section must be first!')
                def_file_format = None
                if section['format']:
                    def_file_format = section['format']
                else:
                    def_file_format = default_format
                def_file.format = def_file_format
                parser = DefFileParser.ManifestParser(section['section'],None,self._get_reader())
                def_file.manifest = parser.parse()
            elif section['dfn_type']:
                parser = self.get_definition_parser(section['dfn_type'],section['section'], section['format'],def_file.format)
                definition = parser.new_section()
                parser.parse(definition)
                def_file.definitions.append(definition)
            else:
                parser = self.get_section_parser(section['section'], section['format'],def_file.format)
                sec = parser.new_section()
                parser.parse(sec)
                def_file.sections[section['section']] = sec
        if not def_file.type and self.default_def_type:
            def_file.type = self.default_def_type
        if self.require_def_type and not re.match(self.require_def_type,def_file.type):
            raise ParsingError("Require def file type {0!s}, got {1!s}".format(self.require_def_type, def_file.type))
        self.fp = None
        self.current_line = 0
        return def_file
                
    class Subparser(object):
        def __init__(self,name,format,reader):
            self.name = name
            self.format = format
            self.reader = reader
        
        def parse(self, section):
            raise NotImplementedError()
    
    class SectionParser(Subparser):
        SECTION_CLASS = None
        
        def new_section(self):
            return self.SECTION_CLASS(self.name)
    
    class DefinitionParser(SectionParser):
        def __init__(self,dfn_type,name,format,reader):
            DefFileParser.SectionParser.__init__(self,name,format,reader)
            self.dfn_type = dfn_type
        
        def new_section(self):
            return self.SECTION_CLASS(self.dfn_type, self.name)
        
        @staticmethod
        def from_section_parser(section_parser):
            if not inspect.isclass(section_parser):
                section_parser = section_parser.__class__
            class DfnParser(section_parser, DefFileParser.DefinitionParser):
                def __init__(self,dfn_type,name,format,reader):
                    DefFileParser.DefinitionParser.__init__(self, dfn_type, name, format, reader)
                    section_parser.__init__(self,name,format,reader)
                
                def new_section(self):
                    return get_definition_from_section(section_parser.SECTION_CLASS)(self.dfn_type, self.name)
            DfnParser.__name__ = _sec_name_to_dfn(section_parser.__name__)
            return DfnParser
    
    class ManifestParser(Subparser):
        def parse(self):
            mf = Manifest()
            for line in self.reader():
                match = DefFileParser.BASIC_KEY_VALUE_RE.match(line)
                if not match:
                    raise ParsingError('Parsing error: invalid manifest line\n{line}'.format(line=line))
                key = match.group('DefFileParser_key')
                value = match.group('DefFileParser_value')
                if key.lower() == 'def-type':
                    mf.def_type = value
                elif key.startswith('include'):
                    dot_loc = key.find('.')
                    if dot_loc == -1:
                        mf.includes.append((value,None))
                    else:
                        include_type = key.split('.')[dot_loc+1:]
                        mf.includes.append((value,include_type))
                else:
                    mf.fields[key] = value
            return mf

def get_definition_parser_from_section_parser(section_parser):
    return DefFileParser.DefinitionParser.from_section_parser(section_parser)

class INISectionParser(DefFileParser.SectionParser):
    SECTION_CLASS = INISection
    KEY_PATT = DefFileParser.BASIC_KEY_PATT
    VALUE_PATT = DefFileParser.BASIC_VALUE_PATT
    
    def parse(self, section):
        line_patt = DefFileParser.KEY_VALUE_PATT_TEMPLATE.format(key_patt=self.KEY_PATT,value_patt=self.VALUE_PATT)
        line_re = re.compile(line_patt)
        
        #section = INISection(self.name)
        
        for line in self.reader():
            match = line_re.match(line)
            if not match:
                raise ParsingError('Parsing error: Invalid INI section line\n{line}'.format(line=line))
            key = re.sub(r'\\=',r'=',match.group('DefFileParser_key'))
            value = match.group('DefFileParser_value')
            section.fields[key] = value
        return section

class ExtendedINISectionParser(INISectionParser):
    KEY_PATT = DefFileParser.BASIC_KEY_PATT
    VALUE_PATT = DefFileParser.BASIC_VALUE_PATT
    EXTENDED_VALUE_PATT = r'.*'
    
    def parse(self, section):
        line_patt = DefFileParser.KEY_VALUE_PATT_TEMPLATE.format(key_patt=self.KEY_PATT,value_patt=self.VALUE_PATT)
        line_re = re.compile(line_patt)
        
        extended_value_re = re.compile(self.EXTENDED_VALUE_PATT)
        
        #section = INISection(self.name)
        
        extended_key = None
        extended_joiner = None
        extended_data = None
        for line in self.reader():
            match = line_re.match(line)
            if extended_key is not None:
                if match:
                    extended_data = re.sub(r'\\=',r'=',extended_joiner.join(extended_data))
                    section.fields[extended_key] = extended_data
                    extended_key = None
                else:
                    extended_match = extended_value_re.match(line)
                    if not extended_match:
                        raise ParsingError('Parsing error: Invalid Extended INI section line for key {key}:\n{line}'.format(key=extended_key,line=line))
                    extended_data.append(line)
                    continue
            if not match:
                raise ParsingError('Parsing error: Invalid Extended INI section line\n{line}'.format(line=line))
            key = re.sub(r'\\=',r'=',match.group('DefFileParser_key'))
            value = match.group('DefFileParser_value')
            if value == '|':
                extended_key = key
                extended_data = []
                extended_joiner = '\n'
            elif value == '>':
                extended_key = key
                extended_data = []
                extended_joiner = ' '
            else:
                extended_key = None
                section.fields[key] = value
        if extended_key is not None:
            extended_data = key = re.sub(r'\\=',r'=',extended_joiner.join(extended_data))
            section.fields[extended_key] = extended_data
        return section

def get_validated_INI_parser(key_patt,value_patt=None,format=None,super_cls=INISectionParser):
    class ValidatedINISectionParser(super_cls):
        if format:
            FORMAT = format
        KEY_PATT = key_patt
        if value_patt is not None:
            VALUE_PATT = value_patt
    return ValidatedINISectionParser

def get_validated_extended_INI_parser(key_patt,value_patt=None,extended_value_patt=None,format=None,super_cls=ExtendedINISectionParser):
    class ValidatedINISectionParser(super_cls):
        if format:
            FORMAT = format
        KEY_PATT = key_patt
        if value_patt is not None:
            VALUE_PATT = value_patt
        if extended_value_patt is not None:
            EXTENDED_VALUE_PATT = extended_value_patt
    return ValidatedINISectionParser

class RawSectionParser(DefFileParser.SectionParser):
    SECTION_CLASS = RawSection
    LINE_PATT = '.*'
    CONTENT_PATT = '.*'
    FOLD = False
    
    def parse(self, section):
        line_re = re.compile(self.LINE_PATT)
        
        #section = RawSection(self.name)
        
        strs = ''
        
        for line in self.reader():
            match = line_re.match(line)
            if not match:
                raise ParsingError('Parsing error: invalid section line\n{line}'.format(line=line))
            strs.append(line)
        joiner = ' ' if self.FOLD else '\n'
        content = joiner.join(strs)
        
        if not re.match(self.CONTENT_PATT,content,flags=re.MULTILINE):
            raise ParsingError('Parsing error: invalid raw section\n{content}'.format(content=content))
        section.content = content
        
        return section

def get_validated_raw_section_parser(line_patt=None,content_patt=None,format=None):
    class ValidatedRawSectionParser(RawSectionParser):
        if format:
            FORMAT = format
        if line_patt:
            LINE_PATT = line_patt
        if content_patt:
            CONTENT_PATT = content_patt
    return ValidatedRawSectionParser

INIDefinitionParser = get_definition_parser_from_section_parser(INISectionParser)
ExtendedINIDefinitionParser = get_definition_parser_from_section_parser(ExtendedINISectionParser)

class ROSStyleDefinitionParser(DefFileParser.DefinitionParser):
    TYPE_PATT = r'\S+'
    NAME_PATT = r'\S+'
    ALLOW_EMPTY_SEGMENTS = True
    
    @classmethod
    def line_patt(cls):
        return r'^(?P<ROSStyleDefinitionParser_type>{type_patt})\s+(?P<ROSStyleDefinitionParser_name>{name_patt})$'.format(type_patt=cls.TYPE_PATT,name_patt=cls.NAME_PATT)
    
    SEGMENT_NAMES = None
    
    def new_section(self):
        return ROSStyleDefinition(self.dfn_type,self.name,self.SEGMENT_NAMES or [])
    
    def parse(self, rosdef):
        line_re = re.compile(self.line_patt())
        
        #rosdef = ROSStyleDefinition(self.dfn_type,self.name,self.SEGMENT_NAMES)
        
        num_segments = len(self.SEGMENT_NAMES or [])
        segs = []
        
        segment = []
        segment_num = 0
        seg_line_change = True
        for line in self.reader():
            if line == '----':
                segs.append(segment)
                segment = []
                segment_num += 1
                seg_line_change = True
                continue
            seg_line_change = False
            match = line_re.match(line)
            if not match:
                raise ParsingError('Parsing error: invalid ROS-style definition line\n{line}'.format(line=line))
            segment.append(ROSStyleDefinition.Field(name=match.group('ROSStyleDefinitionParser_name'),type=match.group('ROSStyleDefinitionParser_type')))
        if seg_line_change:
            if not self.ALLOW_EMPTY_SEGMENTS:
                raise ParsingError('Parsing error: empty segment')
            elif len(segs) == num_segments -1:
                segs.append([])
        else:
            segs.append(segment)
        if num_segments and len(segs) != num_segments:
            raise ParsingError('Parsing error: This ROS-style definition is required to have {0:d} segments: {1!s}'.format(num_segments, self.SEGMENT_NAMES))
        rosdef.segments = tuple(segs)
        if not num_segments:
            rosdef.segment_names = tuple(str(i) for i in xrange(len(segs)))
        return rosdef

def get_ros_style_msg_parser(type_patt=None,name_patt=None,super_cls=ROSStyleDefinitionParser):
    class ROSStyleMsgDefinitionParser(super_cls):
        if type_patt is not None:
            TYPE_PATT = type_patt
        if name_patt is not None:
            NAME_PATT = name_patt
        ALLOW_EMPTY_SEGMENTS = False
        SEGMENT_NAMES = ['msg']
    return ROSStyleMsgDefinitionParser

def get_ros_style_method_parser(type_patt=None,name_patt=None,allow_empty_segments=None,super_cls=ROSStyleDefinitionParser):
    class ROSStyleMethodDefinitionParser(super_cls):
        if type_patt is not None:
            TYPE_PATT = type_patt
        if name_patt is not None:
            NAME_PATT = name_patt
        if allow_empty_segments is not None:
            ALLOW_EMPTY_SEGMENTS = allow_empty_segments
        SEGMENT_NAMES = ['request','response']
    return ROSStyleMethodDefinitionParser

def get_ros_style_action_parser(type_patt=None,name_patt=None,super_cls=ROSStyleDefinitionParser):
    class ROSStyleActionDefinitionParser(super_cls):
        if type_patt is not None:
            TYPE_PATT = type_patt
        if name_patt is not None:
            NAME_PATT = name_patt
        SEGMENT_NAMES = ['goal','result','feedback']
    return ROSStyleActionDefinitionParser

def _get_dicti(name, superclass): 
    class dicti(dict):
        """Dictionary that enables case insensitive searching while preserving case sensitivity 
        when keys are listed, i.e., via keys() or items() methods. 
        
        Works by storing a lowercase version of the key as the new key and stores the original key-value 
        pair as the key's value (values become dictionaries).
        Adjusted from https://gist.github.com/babakness/3901174"""
    
        _kv = namedtuple('kv','key val')
    
        def __init__(self, arg=None, **kwargs):
            self._data = superclass()
            if arg:
                if isinstance(arg, dict):
                    for key, value in arg.iteritems():
                        self.__setitem__(key, value)
                elif isinstance(arg, Iterable):
                    for (key, value) in arg:
                        self.__setitem__(key, value)
            elif kwargs:
                for key, value in kwargs.iteritems():
                    self.__setitem__(key, value)
        
        def __iter__(self):
            return self._data.iteritems()
        
        def __contains__(self, key):
            return self._data.__contains__(key.lower())
      
        def __getitem__(self, key):
            return self._data.__getitem__(key.lower()).val
      
        def __setitem__(self, key, value):
            return self._data.__setitem__(key.lower(), self._kv(key, value))
    
        def get(self, key, default=None):
            try:
                v = self._data.__getitem__(key.lower())
            except KeyError:
                return default
            else:
                return v.val
    
        def has_key(self,key):
            if self._data.get(key):
                return True
            else:
                return False
    
        def items(self):
            return [(v.key, v.val) for v in self._data.itervalues()]
        
        def keys(self):
            return [v.key for v in self._data.itervalues()]
        
        def values(self):
            return [v.val for v in self._data.itervalues()]
        
        def iteritems(self):
            for v in self._data.itervalues():
                yield v.key, v.val
            
        def iterkeys(self):
            for v in self._data.itervalues():
                yield v.key
            
        def itervalues(self):
            for v in self._data.itervalues():
                yield v.val
        
        def copy(self):
            return dict(self.iteritems())
        
        def __str__(self):
            return dict.__str__(self._data)
        
        def __repr__(self):
            return dict.__repr__(self._data)
    dicti.__name__ = name
    return dicti

dicti = _get_dicti('dicti', dict)
ordereddicti = _get_dicti('ordereddicti', OrderedDict)

def _get_strstrdict(name, superclass):
    class strstrdict(superclass):
        def __init__(self, arg=None, **kwargs):
            superclass.__init__(self)
            if arg:
                if isinstance(arg, dict):
                    for key, value in arg.iteritems():
                        self.__setitem__(key, value)
                else:
                    for (key, value) in arg:
                        self.__setitem__(key, value)
            elif kwargs:
                for key, value in kwargs.iteritems():
                    self.__setitem__(key, value)
        
        def __contains__(self, key):
            return superclass.__contains__(self, _clean_string(key))
      
        def __getitem__(self, key):
            return superclass.__getitem__(self, _clean_string(key))
      
        def __setitem__(self, key, value):
            return superclass.__setitem__(self, _clean_string(key), _clean_string(value))
    
        def get(self, key, default=None):
            return superclass.get(self, _clean_string(key), default)
    
        def has_key(self,key):
            return superclass.has_key(self, _clean_string(key))
    strstrdict.__name__ = name
    return strstrdict

strstrdict = _get_strstrdict('strstrdict', dict)
strstrdicti = _get_strstrdict('strstrdicti', dicti)

if __name__ == '__main__':
    data = \
"""[Manifest]
def-type = Application
name = my/application  

[msg:Image]
float val1
int val2

[srv:Reco]
Image image
----
String label
Pose pose

[Services]
reco = Reco
"""
    f = open('test.def','w')
    f.write(data)

"""[Manifest!ros]
def-type = ServiceCollection
name = my/application  

[msg:Image]
float val1
int val2

[srv:Reco]
Image image
----
String label
Pose pose

[Services]
reco_service = RecoService
"""

"""[Manifest]
def-type = Service
type = RecoService
name = reco_service

[msg:Image]
float val1
int val2

[method:Reco]
Image image
----
String label
Pose pose

[Methods]
reco = Reco
"""

"""[Manifest]
def-type = ros:ServiceCollection
type = RecoService
name = reco_service

[msg:Image]
float val1
int val2

[srv:Reco]
Image image
----
String label
Pose pose

[Service]
reco = Reco
"""


# -*- coding: utf-8 -*-
from __future__ import absolute_import

import logging
import os

import sys

from ._version import __version__

# Here in case we want to put logging config in there.
# Useful if pyros started alone as simple "server node".
from . import config

# create logger
_logger = logging.getLogger(__name__)
# and let it propagate to parent logger, or other handler
# the user of pyros should configure handlers

# handling all import trickery to access dependent ROS modules here, our "entry point for import"
try:
    # this should work from python venv or from ROS environment already setup.
    import pkg_resources
    import pyzmp
except (pkg_resources.DistributionNotFound, ImportError):
    # Handling hybrid usecase : package built in a devel workspace with catkin, and used from normal python
    import pyros_setup  # this has to be in your python environment.

    # Lets use the system configuration already setup by pyros-setup
    pyros_setup.configurable_import().configure().activate()
    # This will use default pyros_setup.cfg. see http://flask.pocoo.org/docs/0.11/config/#instance-folders
    # also pyros_setup should create the file if it doesnt exist.

    _logger.warning("Modified Python sys.path {0}".format(sys.path))
    import pyzmp

from .exceptions import PyrosException
from .pyros_ctx_server import pyros_ctx
from .pyros_client import PyrosClient, PyrosServiceNotFound, PyrosServiceTimeout
from .pyros_ros import PyrosROS
from .pyros_mock import PyrosMock

__all__ = [
    '__version__',
    'PyrosException',
    'PyrosServiceNotFound',
    'PyrosServiceTimeout',
    'pyros_ctx',
    'PyrosClient',
    'PyrosROS',
    'PyrosMock',
]


from __future__ import absolute_import
from __future__ import print_function

import socket
import time
import roslaunch


# We wrap rospy function into safeguards for socket error
# since the master seems to be quite sensitive to Network health.
# TODO : refine which ones need which exception handling...


from __future__ import absolute_import

from importlib import import_module

import roslib

from .api import rospy_safe as rospy
from .message_conversion import get_msg, get_msg_dict


def get_topic_msg(topic):
    return get_msg(topic.rostype)


def get_topic_msg_dict(topic):
    return get_msg_dict(topic.rostype)


# Maybe have a global method that generate a context manager to interface with it...
class TopicTuple(object):
    def __init__(self, name, type, endpoints):
        self.name = name
        self.type = type
        self.endpoints = endpoints
# Note : for topic the connection endpoint is important.
# We can have multiple subscribers and publishers, from different node.
# We need to know if we can drop our interface when we receive a difference ( only some pub|sub lost )
# => we need to track endpoints
# TODO: make that the pickled representation of TopicBack (check asdict())


class TopicBase(object):
    """
    TopicBase is the class implementing common behavior between Subscriber and Publisher.
    It should be used to manipulate Topics.
    """

    # We need some kind of instance count here since system state returns only one node instance
    # as publisher for multiple publisher in this process.
    #
    # This is used in ros_interface update for added puba/subs to determine
    # if we previously started publishing / subscribing to this topic
    # usually the count will be just 1, but more is possible during tests
    #
    # This is also used in ros_interface update for removed pubs/subs to determine
    # if we previously stopped publishing / subscribing to this topic
    # usually the count will be just 1, but more is possible during tests

    def __init__(self, topic_name, topic_type, *args, **kwargs):

        topic_name = rospy.resolve_name(topic_name)
        # TODO : simplify this, since we have normalized topic_name in pub and sub interface classes
        self.name = topic_name

        topic_type_module, topic_type_name = tuple(topic_type.split('/'))
        roslib.load_manifest(topic_type_module)
        msg_module = import_module(topic_type_module + '.msg')

        self.rostype_name = topic_type
        self.rostype = getattr(msg_module, topic_type_name)

        self.msgtype = get_topic_msg_dict(self)

    def cleanup(self):
        """
        Launched when we want to whithhold this interface instance
        :return:
        """
        pass

    def asdict(self):
        """
        Here we provide a dictionary suitable for a representation of the Topic instance
        the main point here is to make it possible to transfer this to remote processes.
        We are not interested in pickleing the whole class with Subscriber and Publisher
        :return:
        """
        return {
            'name': self.name,
            'fullname': self.name,  # for BWcompat
            'msgtype': self.msgtype,
            'rostype_name': self.rostype_name,
        }






from __future__ import absolute_import

import logging
import re


# Utility methods to work with regexes
def cap_match_string(match):
    """
    Attach beginning of line and end of line characters to the given string.
    Ensures that raw topics like /test do not match other topics containing
    the same string (e.g. /items/test would result in a regex match at char 7)
    regex goes through each position in the string to find matches - this
    forces it to consider only the first position
    :param match: a regex
    :return:
    """
    return '^' + match + '$'


def find_first_regex_match(key, regex_candidates):
    """
    find the first regex that match with the key
    :param key: a key to try to match against multiple regex
    :param match_candidates: a list of regexes to check if they match the key
    :return: the first candidate found
    """
    for cand in regex_candidates:
        try:
            pattern = re.compile(cap_match_string(cand))
            if pattern.match(key):
                return cand
        except:
            logging.warn('[ros_interface] Ignoring invalid regex string "{0!s}"!'.format(cand))

    return None


def regex_match_sublist(regex, match_candidates):
    """
    Filter the match_candidates list to return only the candidate that match the regex
    :param regex: a regex used to filter the list of candidates
    :param match_candidates: the list of candidates
    :return: the filtered list of only the candidates that match the regex
    """
    matches = []
    try:
        pattern = re.compile(cap_match_string(regex))
        matches = [cand for cand in match_candidates if pattern.match(cand)]
    except:
        logging.warn('[ros_interface] Ignoring invalid regex string "{0!s}"!'.format(regex))
    return matches


def regexes_match_sublist(regexes, match_candidates):
    """
    Filter the match_candidates list to return only the candidate that match the regex
    :param regexes: a list of regex used to filter the list of candidates
    :param match_candidates: the list of candidates
    :return: the filtered list of only the candidates that match the regex
    """

    return [match for sublist in [regex_match_sublist(rgx, match_candidates) for rgx in regexes] for match in sublist]


import os
import subprocess
import sys
import setuptools

# Ref : https://packaging.python.org/single_source_version/#single-sourcing-the-version
with open('pyros/_version.py') as vf:
    exec(vf.read())

# Best Flow :
# Clean previous build & dist
# $ gitchangelog >CHANGELOG.rst
# change version in code and changelog
# $ python setup.py prepare_release
# WAIT FOR TRAVIS CHECKS
# $ python setup.py publish
# => TODO : try to do a simpler "release" command

# TODO : command to retrieve extra ROS stuff from a third party release repo ( for ROS devs ). useful in dev only so maybe "rosdevelop" ? or via catkin_pip ?
# TODO : command to release to Pip and ROS (bloom) same version one after the other...


# Clean way to add a custom "python setup.py <command>"
# Ref setup.py command extension : https://blog.niteoweb.com/setuptools-run-custom-code-in-setup-py/
class PrepareReleaseCommand(setuptools.Command):
    """Command to release this package to Pypi"""
    description = "prepare a release of pyros"
    user_options = []

    def initialize_options(self):
        """init options"""
        pass

    def finalize_options(self):
        """finalize options"""
        pass

    def run(self):
        """runner"""

        # TODO :
        # $ gitchangelog >CHANGELOG.rst
        # $ git commit CHANGELOG.rst -m "updating changelog"
        # change version in code and changelog
        subprocess.check_call("git commit CHANGELOG.rst pyros/_version.py package.xml -m 'v{0}'".format(__version__), shell=True)
        subprocess.check_call("git push", shell=True)

        print("You should verify travis checks, and you can publish this release with :")
        print("  python setup.py publish")
        sys.exit()


# Clean way to add a custom "python setup.py <command>"
# Ref setup.py command extension : https://blog.niteoweb.com/setuptools-run-custom-code-in-setup-py/
class PublishCommand(setuptools.Command):
    """Command to release this package to Pypi"""
    description = "releases pyros to Pypi"
    user_options = []

    def initialize_options(self):
        """init options"""
        pass

    def finalize_options(self):
        """finalize options"""
        pass

    def run(self):
        """runner"""
        # TODO : clean build/ and dist/ before building...
        subprocess.check_call("python setup.py sdist", shell=True)
        subprocess.check_call("python setup.py bdist_wheel", shell=True)
        # OLD way:
        # os.system("python setup.py sdist bdist_wheel upload")
        # NEW way:
        # Ref: https://packaging.python.org/distributing/
        subprocess.check_call("twine upload dist/*", shell=True)

        subprocess.check_call("git tag -a {0} -m 'version {0}'".format(__version__), shell=True)
        subprocess.check_call("git push --tags", shell=True)
        sys.exit()


class ROSPublishCommand(setuptools.Command):
    """Command to release this package to Pypi"""
    description = "releases pyros to ROS"
    user_options = []

    def initialize_options(self):
        """init options"""
        pass

    def finalize_options(self):
        """finalize options"""
        pass

    def run(self):
        """runner"""
        # TODO : distro from parameter. default : ['indigo', 'jade', 'kinetic']
        subprocess.check_call("git tag -a ros-{0} -m 'version {0} for ROS'".format(__version__), shell=True)
        subprocess.check_call("git push --tags", shell=True)

        subprocess.check_call("bloom-release --rosdistro indigo --track indigo pyros", shell=True)
        sys.exit()

setuptools.setup(name='pyros',
    version=__version__,
    description='ROS Node to provide ROS introspection for non-ROS users.',
    url='http://github.com/asmodehn/pyros',
    author='AlexV',
    author_email='asmodehn@gmail.com',
    license='BSD',
    packages=[
        'pyros',
        'pyros.tests',
        'pyros.baseinterface',
        'pyros.rosinterface',
        'pyros.rosinterface.api',
        'pyros.rosinterface.mock',
        'pyros.rosinterface.mock.tests',
        'pyros.rosinterface.tests',
        'pyros.rosinterface.rostests',
        'pyros.zmpinterface',
        'pyros.zmpinterface.tests',
    ],
    entry_points={
        'console_scripts': [
            'pyros = pyros.__main__:nosemain'
        ]
    },
    # this is better than using package data ( since behavior is a bit different from distutils... )
    include_package_data=True,  # use MANIFEST.in during install.
    install_requires=[
        'tblib',  # this might not always install six (latest version does not)
        'six',
        'pyzmq',
        'pyzmp>=0.0.14',  # lets match the requirement in package.xml (greater than)
        'pyros_setup>=0.1.5',  # Careful : pyros-setup < 0.0.8 might already be installed as a deb in /opt/ros/indigo/lib/python2.7/dist-packages/
        'pyros_config>=0.1.4',
        'nose>=1.3.7',
        'mock>=1.0.1',  # old mock to be compatible with trusty versions
    ],
    # Reference for optional dependencies : http://stackoverflow.com/questions/4796936/does-pip-handle-extras-requires-from-setuptools-distribute-based-sources
    test_suite="nose.collector",
    tests_require=["nose"],
    cmdclass={
        'prepare_release': PrepareReleaseCommand,
        'publish': PublishCommand,
    },
    zip_safe=False,  # TODO testing...
)



from __future__ import absolute_import
from __future__ import print_function

import socket
import time
import rospy


# We wrap rospy function into safeguards for socket error
# since the master seems to be quite sensitive to Network health.
# TODO : refine which ones need which exception handling...

retry_timeout = 1  # number of seconds to wait before retrying


def logerr(msg):
    res = None
    while res is None:
        try:
            res = rospy.logerr(msg)
        except (socket.error, socket.herror, socket.gaierror) as e:
            rospy.logerr("Pyros : got socket error calling logerr({msg}). Retrying...".format(**locals()))
            time.sleep(retry_timeout)
        else:
            return res


def loginfo(msg):
    res = None
    while res is None:
        try:
            res = rospy.loginfo(msg)
        except (socket.error, socket.herror, socket.gaierror) as e:
            rospy.logerr("Pyros : got socket error calling loginfo({msg}). Retrying...".format(**locals()))
            time.sleep(retry_timeout)
        else:
            return res


def logwarn(msg):
    res = None
    while res is None:
        try:
            res = rospy.logwarn(msg)
        except (socket.error, socket.herror, socket.gaierror) as e:
            rospy.logerr("Pyros : got socket error calling logwarn({msg}). Retrying...".format(**locals()))
            time.sleep(retry_timeout)
        else:
            return res


def init_node(name, *args, **kwargs):
    res = None
    while res is None:
        try:
            res = rospy.init_node(name, *args, **kwargs)
        except (socket.error, socket.herror, socket.gaierror) as e:
            rospy.logerr("Pyros : got socket error calling init_node({name}, *{args}, **{kwargs}). Retrying...".format(**locals()))
            time.sleep(retry_timeout)
        else:
            return res


def resolve_name(name):
    res = None
    while res is None:
        try:
            res = rospy.resolve_name(name)
        except (socket.error, socket.herror, socket.gaierror) as e:
            rospy.logerr("Pyros : got socket error calling resolve_name({name}). Retrying...".format(**locals()))
            time.sleep(retry_timeout)
        else:
            return res


def get_name():
    res = None
    while res is None:
        try:
            res =rospy.get_name()
        except (socket.error, socket.herror, socket.gaierror) as e:
            rospy.logerr("Pyros : got socket error calling get_name(). Retrying...".format(**locals()))
            time.sleep(retry_timeout)
        else:
            return res
        

def get_param_names():
    res = None
    while res is None:
        try:
            res =rospy.get_param_names()
        except (socket.error, socket.herror, socket.gaierror) as e:
            rospy.logerr("Pyros : got socket error calling get_param_names(). Retrying...".format(**locals()))
            time.sleep(retry_timeout)
        else:
            return res


def get_param(name, default=None):
    res = None
    while res is None:
        try:
            res =rospy.get_param(name, default)
        except (socket.error, socket.herror, socket.gaierror) as e:
            rospy.logerr("Pyros : got socket error calling get_param({name},{default}). Retrying...".format(**locals()))
            time.sleep(retry_timeout)
        else:
            return res
        
        
def set_param(name, value):
    res = None
    while res is None:
        try:
            res = rospy.set_param(name, value)
        except (socket.error, socket.herror, socket.gaierror) as e:
            rospy.logerr("Pyros : got socket error calling set_param({name},{value}). Retrying...".format(**locals()))
            time.sleep(retry_timeout)
        else:
            return res


def delete_param(name):
    res = None
    while res is None:
        try:
            res = rospy.delete_param(name)
        except (socket.error, socket.herror, socket.gaierror) as e:
            rospy.logerr("Pyros : got socket error calling delete_param({name}). Retrying...".format(**locals()))
            time.sleep(retry_timeout)
        else:
            return res
        
        
class MasterAPI_safe(object):
    def __init__(self, ms_proxy):
        self.ms_proxy = ms_proxy

    def lookupNode(self, caller_id):
        res = None
        while res is None:
            try:
                res = self.ms_proxy.lookupNode(caller_id)
            except (socket.error, socket.herror, socket.gaierror) as e:
                rospy.logerr("Pyros : got socket error calling MasterAPI_safe:lookupNode({caller_id}). Retrying...".format(
                    **locals()))
                time.sleep(retry_timeout)
            else:
                return res

    def getSystemState(self):
        res = None
        while res is None:
            try:
                res = self.ms_proxy.getSystemState()
            except (socket.error, socket.herror, socket.gaierror) as e:
                rospy.logerr("Pyros : got socket error calling MasterAPI_safe:getSystemState(). Retrying...".format(**locals()))
                time.sleep(retry_timeout)
            else:
                return res
    
    def getTopicTypes(self):
        res = None
        while res is None:
            try:
                res = self.ms_proxy.getTopicTypes()
            except (socket.error, socket.herror, socket.gaierror) as e:
                rospy.logerr("Pyros : got socket error calling MasterAPI_safe:getTopicTypes(). Retrying...".format(**locals()))
                time.sleep(retry_timeout)
            else:
                return res


def get_master():
    res = None
    while res is None:
        try:
            res = MasterAPI_safe(rospy.get_master())
        except (socket.error, socket.herror, socket.gaierror) as e:
            rospy.logerr(
                "Pyros : got socket error calling get_master(). Retrying...".format(**locals()))
            time.sleep(retry_timeout)
        else:
            return res

# Forwarding class definitions
Subscriber = rospy.Subscriber
Publisher = rospy.Publisher
Service = rospy.Service
ServiceProxy = rospy.ServiceProxy

rostime = rospy.rostime

__all__ = [
    'logerr',
    'loginfo',
    'logwarn',
    'resolve_name',
    'get_param',
    'set_param',
    'delete_param',
    'init_node',
    'get_name',
    'get_param_names',
    'get_master',
    'Publisher',
    'Subscriber',
    'Service',
    'ServiceProxy',
    'rostime',
]


from __future__ import absolute_import
import abc


class BaseMessage(object):

    """
    BaseMsg is a base class to handle message type conversion
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, content_dict):
        """
        This handles conversion between types, the same way that python converts types from one to another
        :param msg:
        :return:
        """
        pass

    def __repr__(self):
        """
        This returns a representation of the message
        :return:
        """
        pass

    def __dict__(self):
        """
        This returns a dict representation of the message.
        Should be the inverse of the constructor
        :return:
        """
        pass




from __future__ import absolute_import

from contextlib import contextmanager

from ...baseinterface import TransientIfPool
from .mocksystem import (
    topics_available_remote, topics_available_type_remote,
)

from .mockpublisher import MockPublisher


class MockPublisherPool(TransientIfPool):

    """
    MockInterface.
    """
    def __init__(self, topics=None):
        # This base constructor assumes the system to interface with is already available ( can do a get_svc_available() )
        super(MockPublisherPool, self).__init__(topics)

    # mock functions that simulate/mock similar interface than what is found on multiprocess framework supported
    # We should try our best to go for the lowest common denominator here
    # TOPICS
    def get_transients_available(self):  # function returning all topics available on the system
        return self.available

    def transient_type_resolver(self, topic_name):  # function resolving the type of a topic
        tpc = self.available.get(topic_name)
        return tpc  # None is returned if not found

    def TransientMaker(self, topic_name, topic_type, *args, **kwargs):  # the topic class implementation
        return MockPublisher(topic_name, topic_type, *args, **kwargs)

    def TransientCleaner(self, topic):  # the topic class implementation
        return topic.cleanup()

    def update(self):
        for t in topics_available_remote.keys():  # explicit keys in case we work through a dictproxy
            # if topics_available_remote.get(t, 0) > 0:  # currently not needed
                self.available[t] = topics_available_type_remote.get(t)

        dt = self.transient_change_detect()

        return dt

TransientIfPool.register(MockPublisherPool)



from __future__ import absolute_import

from contextlib import contextmanager

from ...baseinterface import TransientIfPool
from .mocksystem import (
    params_available_remote, params_available_type_remote,
)

from .mockparam import MockParam


class MockParamPool(TransientIfPool):

    """
    MockInterface.
    """
    def __init__(self, params=None):
        # This base constructor assumes the system to interface with is already available ( can do a get_svc_available() )
        super(MockParamPool, self).__init__(params)

    # mock functions that simulate/mock similar interface than what is found on multiprocess framework supported
    # We should try our best to go for the lowest common denominator here
    # PARAMS
    def get_transients_available(self):  # function returning all params available on the system
        return self.available

    def transient_type_resolver(self, param_name):  # function resolving the type of a param
        prm = self.available.get(param_name)
        return prm  # None is returned if not found

    def TransientMaker(self, param_name, param_type, *args, **kwargs):  # the param class implementation
        return MockParam(param_name, param_type, *args, **kwargs)

    def TransientCleaner(self, param):  # the param class implementation
        return param.cleanup()

    def update(self):
        for p in params_available_remote:
            self.available[p] = params_available_type_remote.get(p)

        dt = self.transient_change_detect()

        return dt

TransientIfPool.register(MockParamPool)
